#
# Test for the PAM responder
#
# Copyright (c) 2018 Red Hat, Inc.
# Author: Sumit Bose <sbose@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Tests for the PAM responder
"""
import os
import stat
import signal
import errno
import subprocess
import time

import config
import intg.ds_openldap
import kdc

import pytest

from .test_files_provider import sync_files_provider
from intg.util import unindent

LDAP_BASE_DN = "dc=example,dc=com"

def provider_list():
     if os.environ['FILES_PROVIDER'] == "enabled":
         return ('files', 'files_with_policy', 'proxy')
     else:
         # The comma is required to indicate a list with the string 'proxy' as
         # only item, without it the string 'proxy' will be interpreted as list
         # with five letters.
         return ('proxy',)


class provider_switch:
    def __init__(self, p):
        if p == 'files':
            self.p = "id_provider = files\nfallback_to_nss = False\n"
        elif p == 'files_with_policy':
            self.p = "id_provider = files\nfallback_to_nss = False\nlocal_auth_policy = enable:smartcard\n"
        elif p == 'proxy':
            self.p = "id_provider = proxy\nlocal_auth_policy = only\nproxy_lib_name = call\n"
        elif p == 'proxy_password':
            self.p = "id_provider = proxy\nproxy_lib_name = call\nproxy_pam_target = sssd-shadowutils\n"
        elif p == 'proxy_password_with_sc':
            self.p = "id_provider = proxy\nlocal_auth_policy = enable:smartcard\nproxy_lib_name = call\nproxy_pam_target = sssd-shadowutils\n"
        else:
            self.p = none


@pytest.fixture(scope="module")
def ad_inst(request):
    """Fake AD server instance fixture"""
    instance = intg.ds_openldap.FakeAD(
        config.PREFIX, 10389, LDAP_BASE_DN,
        "cn=admin", "Secret123"
    )

    try:
        instance.setup()
    except Exception:
        instance.teardown()
        raise
    request.addfinalizer(instance.teardown)
    return instance


@pytest.fixture(scope="module")
def ldap_conn(request, ad_inst):
    """LDAP server connection fixture"""
    ldap_conn = ad_inst.bind()
    ldap_conn.ad_inst = ad_inst
    request.addfinalizer(ldap_conn.unbind_s)
    return ldap_conn


def format_basic_conf(ldap_conn):
    """Format a basic SSSD configuration"""
    return unindent("""\
        [sssd]
        domains = FakeAD
        services = pam, nss

        [nss]

        [pam]
        debug_level = 10

        [domain/FakeAD]
        debug_level = 10
        ldap_search_base = {ldap_conn.ad_inst.base_dn}
        ldap_referrals = false

        id_provider = ldap
        auth_provider = ldap
        chpass_provider = ldap
        access_provider = ldap

        ldap_uri = {ldap_conn.ad_inst.ldap_url}
        ldap_default_bind_dn = {ldap_conn.ad_inst.admin_dn}
        ldap_default_authtok_type = password
        ldap_default_authtok = {ldap_conn.ad_inst.admin_pw}

        ldap_schema = ad
        ldap_id_mapping = true
        ldap_idmap_default_domain_sid = S-1-5-21-1305200397-2901131868-73388776
        case_sensitive = False

        [prompting/password]
        password_prompt = My global prompt

        [prompting/password/pam_sss_alt_service]
        password_prompt = My alt service prompt
    """).format(**locals())

USER1 = dict(name='user1', passwd='x', uid=10001, gid=20001,
             gecos='User for tests',
             dir='/home/user1',
             shell='/bin/bash')

USER2 = dict(name='user2', passwd='x', uid=10002, gid=20002,
             gecos='User with no Smartcard mapping',
             dir='/home/user2',
             shell='/bin/bash')


def format_pam_cert_auth_conf(config, provider):
    """Format a basic SSSD configuration"""
    return unindent("""\
        [sssd]
        debug_level = 10
        domains = auth_only
        services = pam, nss

        [nss]
        debug_level = 10

        [pam]
        pam_cert_auth = True
        pam_p11_allowed_services = +pam_sss_service, +pam_sss_sc_required, \
                                   +pam_sss_try_sc, +pam_sss_allow_missing_name
        pam_cert_db_path = {config.PAM_CERT_DB_PATH}
        p11_uri = pkcs11:manufacturer=SoftHSM%20project; \
                  token=SSSD%20Test%20Token
        p11_child_timeout = 5
        p11_wait_for_card_timeout = 5
        debug_level = 10

        [domain/auth_only]
        debug_level = 10
        {provider.p}

        [certmap/auth_only/user1]
        matchrule = <SUBJECT>.*CN=SSSD test cert 000[12].*
    """).format(**locals())


def format_pam_cert_auth_conf_name_format(config, provider):
    """Format SSSD configuration with full_name_format"""
    return unindent("""\
        [sssd]
        debug_level = 10
        domains = auth_only
        services = pam, nss

        [nss]
        debug_level = 10

        [pam]
        pam_cert_auth = True
        pam_p11_allowed_services = +pam_sss_service, +pam_sss_sc_required, \
                                   +pam_sss_try_sc, +pam_sss_allow_missing_name
        pam_cert_db_path = {config.PAM_CERT_DB_PATH}
        p11_uri = pkcs11:manufacturer=SoftHSM%20project; \
                  token=SSSD%20Test%20Token
        p11_child_timeout = 5
        p11_wait_for_card_timeout = 5
        debug_level = 10

        [domain/auth_only]
        use_fully_qualified_names = True
        full_name_format = %2$s\\%1$s
        debug_level = 10
        {provider.p}

        [certmap/auth_only/user1]
        matchrule = <SUBJECT>.*CN=SSSD test cert 000[12].*
    """).format(**locals())


def format_pam_krb5_auth(config, kdc_instance):
    """Format SSSD configuration for krb5 authentication"""
    return unindent("""\
        [sssd]
        debug_level = 10
        domains = krb5_auth
        services = pam, nss

        [nss]
        debug_level = 10

        [pam]
        debug_level = 10

        [domain/krb5_auth]
        debug_level = 10
        id_provider = proxy
        proxy_lib_name = call
        auth_provider = krb5

        krb5_realm = PAMKRB5TEST
        krb5_server = localhost:{kdc_instance.kdc_port}
    """).format(**locals())


def format_pam_krb5_auth_domains(config, kdc_instance):
    """Format SSSD configuration for krb5 authentication"""
    return unindent("""\
        [sssd]
        debug_level = 10
        domains = wrong.dom1, wrong.dom2, krb5_auth, wrong.dom3
        services = pam, nss

        [nss]
        debug_level = 10

        [pam]
        debug_level = 10

        [domain/wrong.dom1]
        debug_level = 10
        id_provider = proxy
        proxy_lib_name = call
        auth_provider = krb5

        krb5_realm = WRONG1REALM
        krb5_server = localhost:{kdc_instance.kdc_port}

        [domain/wrong.dom2]
        debug_level = 10
        id_provider = proxy
        proxy_lib_name = call
        auth_provider = krb5

        krb5_realm = WRONG2REALM
        krb5_server = localhost:{kdc_instance.kdc_port}

        [domain/wrong.dom3]
        debug_level = 10
        id_provider = proxy
        proxy_lib_name = call
        auth_provider = krb5

        krb5_realm = WRONG3REALM
        krb5_server = localhost:{kdc_instance.kdc_port}

        [domain/krb5_auth]
        debug_level = 10
        id_provider = proxy
        proxy_lib_name = call
        auth_provider = krb5

        krb5_realm = PAMKRB5TEST
        krb5_server = localhost:{kdc_instance.kdc_port}
    """).format(**locals())


def create_conf_file(contents):
    """Create sssd.conf with specified contents"""
    conf = open(config.CONF_PATH, "w")
    conf.write(contents)
    conf.close()
    os.chmod(config.CONF_PATH, stat.S_IRUSR | stat.S_IWUSR)


def create_conf_fixture(request, contents):
    """
    Create sssd.conf with specified contents and add teardown for removing it
    """
    create_conf_file(contents)

    def cleanup_conf_file():
        """Remove sssd.conf, if it exists"""
        if os.path.lexists(config.CONF_PATH):
            os.unlink(config.CONF_PATH)

    request.addfinalizer(cleanup_conf_file)


def create_sssd_process(krb5_conf_path=None):
    """Start the SSSD process"""
    my_env = os.environ.copy()
    my_env["SSS_FILES_PASSWD"] = os.environ["NSS_WRAPPER_PASSWD"]
    my_env["SSS_FILES_GROUP"] = os.environ["NSS_WRAPPER_GROUP"]
    my_env['SSSD_INTG_PEER_UID'] = "0"
    my_env['SSSD_INTG_PEER_GID'] = "0"
    if krb5_conf_path is not None:
        my_env['KRB5_CONFIG'] = krb5_conf_path
    if subprocess.call(["sssd", "-D", "--logger=files"], env=my_env) != 0:
        raise Exception("sssd start failed")


def cleanup_sssd_process():
    """Stop the SSSD process and remove its state"""
    try:
        with open(config.PIDFILE_PATH, "r") as pid_file:
            pid = int(pid_file.read())
        os.kill(pid, signal.SIGTERM)
        while True:
            try:
                os.kill(pid, signal.SIGCONT)
            except OSError:
                break
            time.sleep(1)
    except OSError:
        pass
    for path in os.listdir(config.DB_PATH):
        os.unlink(config.DB_PATH + "/" + path)
    for path in os.listdir(config.MCACHE_PATH):
        os.unlink(config.MCACHE_PATH + "/" + path)

    # make sure that the indicator file is removed during shutdown
    try:
        assert not os.stat(config.PUBCONF_PATH + "/pam_preauth_available")
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise ex


def create_sssd_fixture(request, krb5_conf_path=None):
    """Start SSSD and add teardown for stopping it and removing its state"""
    create_sssd_process(krb5_conf_path)
    request.addfinalizer(cleanup_sssd_process)


@pytest.fixture
def simple_pam_cert_auth(request, passwd_ops_setup):
    """Setup SSSD with pam_cert_auth=True"""
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']
    conf = format_pam_cert_auth_conf(config, provider_switch(request.param))
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    sync_files_provider(USER2['name'])

    return None


@pytest.fixture
def simple_pam_cert_auth_no_cert(request, passwd_ops_setup):
    """Setup SSSD with pam_cert_auth=True"""
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']

    old_softhsm2_conf = os.environ['SOFTHSM2_CONF']
    del os.environ['SOFTHSM2_CONF']

    conf = format_pam_cert_auth_conf(config, provider_switch(request.param))
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)

    os.environ['SOFTHSM2_CONF'] = old_softhsm2_conf

    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    sync_files_provider(USER2['name'])

    return None


@pytest.fixture
def simple_pam_cert_auth_two_certs(request, passwd_ops_setup):
    """Setup SSSD with pam_cert_auth=True"""
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']

    old_softhsm2_conf = os.environ['SOFTHSM2_CONF']
    softhsm2_two_conf = os.environ['SOFTHSM2_TWO_CONF']
    os.environ['SOFTHSM2_CONF'] = softhsm2_two_conf

    conf = format_pam_cert_auth_conf(config, provider_switch(request.param))
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)

    os.environ['SOFTHSM2_CONF'] = old_softhsm2_conf

    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    sync_files_provider(USER2['name'])

    return None


@pytest.fixture
def simple_pam_cert_auth_name_format(request, passwd_ops_setup):
    """Setup SSSD with pam_cert_auth=True and full_name_format"""
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']
    conf = format_pam_cert_auth_conf_name_format(config, provider_switch(request.param))
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    sync_files_provider(f"{USER2['name']}@auth_only")

    return None

@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_preauth_indicator(simple_pam_cert_auth):
    """Check if preauth indicator file is created"""
    statinfo = os.stat(config.PUBCONF_PATH + "/pam_preauth_available")
    assert stat.S_ISREG(statinfo.st_mode)


@pytest.fixture
def pam_prompting_config(request, ldap_conn):
    """Setup SSSD with PAM prompting config"""
    conf = format_basic_conf(ldap_conn)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_password_prompting_config_global(ldap_conn, pam_prompting_config,
                                          env_for_sssctl):
    """Check global change of the password prompt"""

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1_dom1-19661",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="111")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("My global prompt") != -1


def test_password_prompting_config_srv(ldap_conn, pam_prompting_config,
                                       env_for_sssctl):
    """Check change of the password prompt for dedicated service"""

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1_dom1-19661",
                               "--action=auth",
                               "--service=pam_sss_alt_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="111")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("My alt service prompt") != -1


@pytest.fixture
def env_for_sssctl(request):
    pwrap_runtimedir = os.getenv("PAM_WRAPPER_SERVICE_DIR")
    if pwrap_runtimedir is None:
        raise ValueError("The PAM_WRAPPER_SERVICE_DIR variable is unset\n")

    env_for_sssctl = os.environ.copy()
    env_for_sssctl['PAM_WRAPPER'] = "1"
    env_for_sssctl['SSSD_INTG_PEER_UID'] = "0"
    env_for_sssctl['SSSD_INTG_PEER_GID'] = "0"
    env_for_sssctl['LD_PRELOAD'] += ':' + os.environ['PAM_WRAPPER_PATH']

    return env_for_sssctl


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_sc_auth_wrong_pin(simple_pam_cert_auth, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="111")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: "
                    "Authentication failure") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_sc_auth(simple_pam_cert_auth, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth_two_certs', provider_list(), indirect=True)
def test_sc_auth_two(simple_pam_cert_auth_two_certs, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="2\n123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth_two_certs', provider_list(), indirect=True)
def test_sc_auth_two_missing_name(simple_pam_cert_auth_two_certs, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "",
                               "--action=auth", "--service=pam_sss_allow_missing_name"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="2\n123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', ['proxy_password'], indirect=True)
def test_sc_proxy_password_fallback(simple_pam_cert_auth, env_for_sssctl):
    """
    Check that there will be a password prompt if another proxy auth module is
    configured and Smartcard authentication is not allowed but a Smartcard is
    present.
    """

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    assert err.find("Password:") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', ['proxy_password_with_sc'],
                         indirect=True)
def test_sc_proxy_no_password_fallback(simple_pam_cert_auth, env_for_sssctl):
    """
    Use the same environ as for test_sc_proxy_password_fallback but now allow
    local Smartcard authentication. Here we expect that there will be a prompt
    for the Smartcard PIN and that Smartcard authentication is successful.
    """

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth", "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_require_sc_auth(simple_pam_cert_auth, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth",
                               "--service=pam_sss_sc_required"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth_no_cert', provider_list(), indirect=True)
def test_require_sc_auth_no_cert(simple_pam_cert_auth_no_cert, env_for_sssctl):

    # We have to wait about 20s before the command returns because there will
    # be 2 run since retry=1 in the PAM configuration and both
    # p11_child_timeout and p11_wait_for_card_timeout are 5s in sssd.conf,
    # so 2*(5+5)=20. */
    start_time = time.time()
    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth",
                               "--service=pam_sss_sc_required"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    end_time = time.time()
    assert end_time > start_time and \
        (end_time - start_time) >= 20 and \
        (end_time - start_time) < 40
    assert out.find("Please insert smart card\nPlease insert smart card") != -1
    assert err.find("pam_authenticate for user [user1]: Authentication "
                    "service cannot retrieve authentication info") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_try_sc_auth_no_map(simple_pam_cert_auth, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user2",
                               "--action=auth",
                               "--service=pam_sss_try_sc"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user2]: Authentication "
                    "service cannot retrieve authentication info") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_try_sc_auth(simple_pam_cert_auth, env_for_sssctl):

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth",
                               "--service=pam_sss_try_sc"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_try_sc_auth_root(simple_pam_cert_auth, env_for_sssctl):
    """
    Make sure pam_sss returns PAM_AUTHINFO_UNAVAIL even for root if
    try_cert_auth is set.
    """
    sssctl = subprocess.Popen(["sssctl", "user-checks", "root",
                               "--action=auth",
                               "--service=pam_sss_try_sc"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [root]: Authentication "
                    "service cannot retrieve authentication info") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_sc_auth_missing_name(simple_pam_cert_auth, env_for_sssctl):
    """
    Test pam_sss allow_missing_name feature.
    """

    sssctl = subprocess.Popen(["sssctl", "user-checks", "",
                               "--action=auth",
                               "--service=pam_sss_allow_missing_name"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth', provider_list(), indirect=True)
def test_sc_auth_missing_name_whitespace(simple_pam_cert_auth, env_for_sssctl):
    """
    Test pam_sss allow_missing_name feature.
    """

    sssctl = subprocess.Popen(["sssctl", "user-checks", " ",
                               "--action=auth",
                               "--service=pam_sss_allow_missing_name"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find("pam_authenticate for user [user1]: Success") != -1


@pytest.mark.parametrize('simple_pam_cert_auth_name_format', provider_list(), indirect=True)
def test_sc_auth_name_format(simple_pam_cert_auth_name_format, env_for_sssctl):
    """
    Test that full_name_format is respected with pam_sss allow_missing_name
    option.
    """

    sssctl = subprocess.Popen(["sssctl", "user-checks", "",
                               "--action=auth",
                               "--service=pam_sss_allow_missing_name"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="123456")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find(r"pam_authenticate for user [auth_only\user1]: "
                    "Success") != -1


@pytest.fixture
def kdc_instance(request):
    """Kerberos server instance fixture"""
    kdc_instance = kdc.KDC(config.PREFIX, "PAMKRB5TEST")
    try:
        kdc_instance.set_up()
        kdc_instance.start_kdc()
    except Exception:
        kdc_instance.teardown()
        raise
    request.addfinalizer(kdc_instance.teardown)
    return kdc_instance


@pytest.fixture
def setup_krb5(request, kdc_instance, passwd_ops_setup):
    """
    Setup SSSD for Kerberos authentication with 2 users with different
    passwords
    """
    conf = format_pam_krb5_auth(config, kdc_instance)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request, kdc_instance.krb5_conf_path)

    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    kdc_instance.add_principal("user1", "Secret123User1")
    kdc_instance.add_principal("user2", "Secret123User2")
    time.sleep(2)   # Give KDC time to initialize
    return None


def test_krb5_auth(setup_krb5, env_for_sssctl):
    """
    Test basic Kerberos authentication, check for authentication failure when
    a wrong password is used
    """
    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth",
                               "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="Secret123User1")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find(r"pam_authenticate for user [user1]: Success") != -1

    sssctl = subprocess.Popen(["sssctl", "user-checks", "user2",
                               "--action=auth",
                               "--service=pam_sss_service"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="Secret123User1")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find(r"pam_authenticate for user [user2]: "
                    "Authentication failure") != -1


@pytest.fixture
def setup_krb5_domains(request, kdc_instance, passwd_ops_setup):
    """
    Setup SSSD for Kerberos authentication with 2 users with different
    passwords and multiple domains configured in sssd.conf
    """
    conf = format_pam_krb5_auth_domains(config, kdc_instance)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request, kdc_instance.krb5_conf_path)

    passwd_ops_setup.useradd(**USER1)
    passwd_ops_setup.useradd(**USER2)
    kdc_instance.add_principal("user1", "Secret123User1")
    kdc_instance.add_principal("user2", "Secret123User2")
    return None


def test_krb5_auth_domains(setup_krb5_domains, env_for_sssctl):
    """
    Test basic Kerberos authentication with pam_sss 'domains' option, make
    sure not-matching domains are skipped even if the user exists in that
    domain
    """
    sssctl = subprocess.Popen(["sssctl", "user-checks", "user1",
                               "--action=auth",
                               "--service=pam_sss_domains"],
                              universal_newlines=True,
                              env=env_for_sssctl, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        out, err = sssctl.communicate(input="Secret123User1")
    except Exception:
        sssctl.kill()
        out, err = sssctl.communicate()

    sssctl.stdin.close()
    sssctl.stdout.close()

    if sssctl.wait() != 0:
        raise Exception("sssctl failed")

    assert err.find(r"pam_authenticate for user [user1]: Success") != -1



""" Automation of sssd nss suite

:requirement: IDM-SSSD-REQ : Configuration and Service Management
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import time
import os
import re
import pytest
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name


def execute_cmd(multihost, command):
    """ Execute command on client """
    cmd = multihost.client[0].run_command(command)
    return cmd


@pytest.mark.usefixtures('setup_sssd')
@pytest.mark.tier1_3
class TestSssdNss(object):
    """
    This is test case class for sssd nss suite
    """
    def test_avoid_interlocking_among_threads(self, multihost,
                                              backupsssdconf):
        """
        :title: avoid interlocking among threads that use
          `libsss_nss_idmap` API (or other sss_client libs)
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1978119
        :id: 836759ee-fc19-11ec-b57d-845cf3eff344
        :customerscenario: true
        :steps:
            1. Configure sssd with nss
            2. Create multiple tread from single process
            3. Check that requests are not being handled serially
        :expectedresults:
            1. Should succeed
            2. Should succeed
            3. Should succeed
        """
        version_fedora = 0
        version_rel_cent = 0
        if "Fedora" in multihost.client[0].distro:
            version_fedora = int(re.search(r'\d+', multihost.client[0].distro).group())
        else:
            version_rel_cent = float(re.findall(r"\d+\.\d+", multihost.client[0].distro)[0])
        if version_rel_cent < 9 and version_fedora < 35:
            pytest.skip("unsupported configuration")
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        services = {'debug_level': '9'}
        tools.sssd_conf('nss', services)
        tools.clear_sssd_cache()
        client = multihost.client[0]
        client.run_command("yum install -y gcc")
        file_location = '/script/thread.c'
        client.transport.put_file(os.path.dirname(os.path.abspath(__file__))
                                  + file_location,
                                  '/tmp/thread.c')
        execute_cmd(multihost, "gcc -lpthread /tmp/thread.c -o /tmp/thread")
        execute_cmd(multihost, ">/var/log/sssd/sssd_nss.log")
        execute_cmd(multihost, "chmod 755 /tmp/thread")
        execute_cmd(multihost, "cd /tmp/; ./thread")
        time.sleep(2)
        cmd = execute_cmd(multihost, "grep 'Looking up' "
                                     "/var/log/sssd/sssd_nss.log").stdout_text
        crs = [int(i.split('#')[1]) for i in re.findall('CR #[0-9]+', cmd)]
        assert sorted(crs) != crs



from __future__ import annotations
import os
from collections import defaultdict

import pytest


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: pytest.Config) -> None:
    p = CustomLogPlugin(config)
    config.pluginmanager.register(p, 'custom_log_plugin')


class CustomLogPlugin:
    """
    Custom log class to dulicate or remove log from console
    and place it inside per-test log files
    """
    def __init__(self, config: pytest.Config) -> None:
        self.config: pytest.Config = config
        self.log_per_test: str = config.getoption("log_per_test")
        self.tests = defaultdict(dict)
        self.backup_showlocals = self.config.option.showlocals
        self.backup_reportchars = self.config.option.reportchars
        self.backup_tbstyle = self.config.option.tbstyle
        self.backup_showcapture = self.config.option.showcapture
        self.backup_capture = self.config.option.capture
        print(f"Custom log settings: '{self.log_per_test}'.")

    def _write_log(self, test: str, phases: list = None) -> None:
        """
        Write log for test, possibly only selected
        phases can be written.
        """
        if not phases:
            _phases = ["setup", "call", "teardown"]
        else:
            _phases = phases
        tr = self.tests[test]
        test_name = test.split("::")[-1]
        test_name = test_name.translate(
            str.maketrans('":<>|*? [/', "----------", "]()"))
        if self.config.option.log_file:
            logdir = os.path.join(os.path.dirname(self.config.option.log_file), 'logs')
        else:
            logdir = os.path.join(os.getcwd(), 'logs')
        os.makedirs(logdir, exist_ok=True)
        logpath = os.path.join(logdir, f'{test_name}.log')
        with open(logpath, 'a+') as f:
            for phase in _phases:
                if phase not in tr:
                    # We do not fail on missing phase as 'call'
                    # could be missing when setup failed.
                    continue
                skip_out = skip_err = skip_log = 0
                if phase == 'call' and 'setup' in tr:
                    skip_out = len(tr['setup'].capstdout)
                    skip_err = len(tr['setup'].capstderr)
                    skip_log = len(tr['setup'].caplog)
                elif phase == 'teardown':
                    if 'setup' in tr:
                        skip_out += len(tr['setup'].capstdout)
                        skip_err += len(tr['setup'].capstderr)
                        skip_log += len(tr['setup'].caplog)
                    if 'call' in tr:
                        skip_out += len(tr['call'].capstdout)
                        skip_err += len(tr['call'].capstderr)
                        skip_log += len(tr['call'].caplog)

                f.write(f"\nPHASE: {phase.upper()} for {test_name}"
                        f"... [{tr[phase].outcome.upper()}]\n")
                if tr[phase].capstdout:
                    f.write(f"\n=== {test_name} {phase.upper()} OUT ===\n")
                    f.write(tr[phase].capstdout[skip_out:])
                if tr[phase].capstderr:
                    f.write(f"\n=== {test_name} {phase.upper()} ERR ===\n")
                    f.write(tr[phase].capstderr[skip_err:])
                if tr[phase].caplog:
                    f.write(f"\n=== {test_name} {phase.upper()} LOG ===\n")
                    f.write(tr[phase].caplog[skip_log:])
                if tr[phase].longreprtext:
                    f.write(f"'\n=== {test_name} {phase.upper()} INFO ===\n")
                    f.write(tr[phase].longreprtext)

    def pytest_runtest_logstart(self, nodeid, location) -> None:
        """Reduce logging to the console when test starts."""
        if self.log_per_test not in ["never", "duplicate"]:
            self.config.option.showlocals = True
            self.config.option.reportchars = 'a'
            self.config.option.tbstyle = 'long'
            self.config.option.showcapture = 'no'
            self.config.option.capture = 'fd'
        else:
            self.config.option.capture = 'tee-sys'

    def pytest_runtest_logfinish(self, nodeid, location) -> None:
        """Revert logging to the console when test ends."""
        if self.log_per_test not in ["never", "duplicate"]:
            self.config.option.showlocals = self.backup_showlocals
            self.config.option.reportchars = self.backup_reportchars
            self.config.option.tbstyle = self.backup_tbstyle
            self.config.option.showcapture = self.backup_showcapture
        self.config.option.capture = self.backup_capture

    def pytest_runtest_logreport(self, report: pytest.TestReport) -> None:
        """
        Hook called on finished test setup, call and teardown
        """
        self.tests[report.nodeid][report.when] = report
        if self.log_per_test == 'always':
            # When we write log always we can write it for each phase
            # This might help when some phase gets stuck.
            self._write_log(report.nodeid, phases=[report.when])
        elif report.when == 'teardown':
            # When writing on-failure we need to wait for teardown to decide
            test = self.tests[report.nodeid]
            if test['setup'].outcome == 'failed' or\
                    test['call'].outcome == 'failed' or\
                    test['teardown'].outcome == 'failed':
                self._write_log(report.nodeid)


def pytest_addoption(parser: pytest.Parser) -> None:
    """
    Pytest hook: add command line options.
    """
    parser.addoption(
        "--log-per-test",
        action="store",
        default="on-failure",
        nargs="?",
        choices=["never", "on-failure", "always", "duplicate"],
        help="Create per-test logfile. 'never': Log everything to console. "
             "'duplicate': Log to console and on-failure also to file. "
             "'always': Keep console clean and always log to file. "
             "'on-failure': Keep console clean and log to file only on "
             "a failure. (default: %(default)s)",
    )



"""
SSSD Authentication Test Cases

:requirement: authentication
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.ad import AD
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.importance("critical")
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_authentication__with_default_settings(
    client: Client, provider: GenericProvider, method: str, sssd_service_user: str
):
    """
    :title: Authenticate with default settings
    :setup:
        1. Create user
        2. Start SSSD
    :steps:
        1. Authenticate user with correct password
        2. Authenticate user with incorrect password
    :expectedresults:
        1. Authentication is successful
        2. Authentication is unsuccessful
    :customerscenario: False
    """
    provider.user("user1").add(password="Secret123")
    if method == "ssh" and "ssh" not in client.sssd.sssd["services"]:
        client.sssd.sssd["services"] = "nss, pam, ssh"
    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.parametrize(method).password("user1", "Secret123"), "User failed login!"
    assert not client.auth.parametrize(method).password(
        "user1", "NOTSecret123"
    ), "User logged in with an invalid password!"


@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.importance("critical")
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_authentication__default_settings_when_the_provider_is_offline(
    client: Client, provider: GenericProvider, method: str, sssd_service_user: str
):
    """
    :title: Authenticate with default settings when the provider is offline
    :setup:
        1. Create user
        2. Configure SSSD with "cache_credentials = true" and "krb5_store_password_if_offline = true" and
        "offline_credentials_expiration = 0"
        3 Start SSSD
    :steps:
        1. Authenticate user with correct password
        2. Offline user authentication with correct password
        3. Offline user authentication with incorrect password
    :expectedresults:
        1. User authentication is successful
        2. User authentication is successful
        3. User authentication is unsuccessful
    :customerscenario: False
    """
    user = "user1"
    correct = "Secret123"
    wrong = "Wrong123"
    provider.user(user).add(password=correct)
    if method == "ssh" and "ssh" not in client.sssd.sssd["services"]:
        client.sssd.sssd["services"] = "nss, pam, ssh"
    client.sssd.domain["cache_credentials"] = "True"
    client.sssd.domain["krb5_store_password_if_offline"] = "True"
    client.sssd.pam["offline_credentials_expiration"] = "0"
    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.parametrize(method).password(user, correct), "User failed login!"

    client.firewall.outbound.reject_host(provider)

    # There might be active connections that are not terminated by creating firewall rule.
    # We need to terminate it by forcing SSSD offline.
    client.sssd.bring_offline()

    assert client.auth.parametrize(method).password(user, correct), "User failed login!"
    assert not client.auth.parametrize(method).password(user, wrong), "User logged in with an incorrect password!"


@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.ticket(gh=7174)
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.importance("critical")
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_authentication__using_the_users_email_address(client: Client, ad: AD, method: str, sssd_service_user: str):
    """
    :title: Login using the user's email address
    :description:
        Testing the feature to login using an email address instead of the userid. The username used,
        must match one of the user's LDAP attribute values, "EmailAddress". The login should be
        case-insensitive and permit special characters.
    :setup:
        1. Add AD users with different email addresses
        2. Start SSSD
    :steps:
        1. Authenticate users using their email address and in different cases
    :expectedresults:
        1. Authentication is successful using the email address and is case-insensitive
    :customerscenario: False
    """
    ad.user("user-1").add(password="Secret123", email=f"user-1@{ad.host.domain}")
    ad.user("user-2").add(password="Secret123", email="user-2@alias-domain.com")
    ad.user("user_3").add(password="Secret123", email="user_3@alias-domain.com")

    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.parametrize(method).password(
        f"user-1@{ad.host.domain}", "Secret123"
    ), f"User user-1@{ad.host.domain} failed login!"
    assert client.auth.parametrize(method).password(
        "user-2@alias-domain.com", "Secret123"
    ), "User user-2@alias-domain.com failed login!"
    assert client.auth.parametrize(method).password(
        "uSEr_3@alias-dOMain.com", "Secret123"
    ), "User uSEr_3@alias-dOMain.com failed login!"




""" Common AD Fixtures """
from __future__ import print_function
import random
import subprocess
import time
import pytest
import os
import posixpath
import tempfile
from sssd.testlib.common.paths import SSSD_DEFAULT_CONF
from sssd.testlib.common.exceptions import SSSDException
from sssd.testlib.common.utils import ADOperations
from sssd.testlib.common.samba import sambaTools
from sssd.testlib.common.utils import sssdTools

pytest_plugins = (
    'sssd.testlib.common.fixtures',
    'pytest_importance',
    'pytest_ticket',
    'sssd.testlib.common.custom_log',
)


def pytest_configure():
    """ Namespace hook, Adds below dict to pytest namespace """
    pytest.num_masters = 1
    pytest.num_ad = 1
    pytest.num_atomic = 0
    pytest.num_replicas = 0
    pytest.num_clients = 1
    pytest.num_others = 0

# ######## Helper functions ####################


def _adleave(client):
    """ Disjoin AD
    @param client: sssdTools instance of the client
    """
    client.disjoin_ad()
    client.multihost.run_command('kdestroy -A')
    # Keytab should be deleted after realm leave
    client.multihost.run_command('rm -f /etc/krb5.keytab', raiseonerr=False)


# ######## Function scoped Fixtures ####################


@pytest.fixture(scope="function")
def smbconfig(session_multihost, request):
    """ Configure smb.conf """
    sambaclient = sambaTools(session_multihost.client[0],
                             session_multihost.ad[0])
    sambaclient.smbadsconf()

    def restore():
        """ Restore smb.conf """
        restoresmb = 'cp -f /etc/samba/smb.conf.orig /etc/samba/smb.conf'
        session_multihost.client[0].run_command(restoresmb, raiseonerr=False)
        removebkup = 'rm -f /etc/samba/smb.conf.orig'
        session_multihost.client[0].run_command(removebkup, raiseonerr=False)
    request.addfinalizer(restore)


@pytest.fixture(scope='function')
def run_powershell_script(session_multihost, request):
    """ Run Powershell script """
    cwd = os.path.dirname(os.path.abspath(__file__))
    split_cwd = cwd.split('/')
    idx = split_cwd.index('multihost')
    path_list = split_cwd[:idx + 1]
    sssd_qe_path = '/'.join(path_list)
    data_path = "%s/data" % sssd_qe_path

    def _script(name):
        """ Run powershell script """
        filename = name
        remote_file_path = posixpath.join('/home/administrator', filename)
        source_file_path = posixpath.join(data_path, filename)
        session_multihost.ad[0].transport.put_file(source_file_path,
                                                   remote_file_path)
        pwrshell_cmd = 'powershell.exe -inputformat '\
                       'none -noprofile ./%s' % filename
        cmd = session_multihost.ad[0].run_command(pwrshell_cmd,
                                                  raiseonerr=False)
        return cmd
    return _script


@pytest.fixture(scope="function")
def create_adgrp(session_multihost, request, run_powershell_script):
    """ fixture to create AD Groups using powershell """
    ret = run_powershell_script(name='adgroup.ps1')
    if ret.returncode == 0:
        grp = ret.stdout_text.strip()
    else:
        raise SSSDException("powershell script failed to execute")

    def delete_adgrp():
        """ Delete AD group """
        dn_entry = session_multihost.ad[0].domain_basedn_entry
        grp_dn_entry = '{},{}'.format('CN=users', dn_entry)
        ad_group_dn = 'cn={},{}'.format(grp, grp_dn_entry)
        dsrm_exe = "dsrm.exe %s -noprompt" % ad_group_dn
        session_multihost.ad[0].run_command(dsrm_exe)
        rm_ps = "rm adgroup.ps1"
        session_multihost.ad[0].run_command(rm_ps, raiseonerr=False)
    request.addfinalizer(delete_adgrp)


@pytest.fixture(autouse=True)
def capture_sssd_logs(session_multihost, request):
    """This will print sssd logs in case of test failure"""
    yield
    if request.session.testsfailed:
        client = session_multihost.client[0]
        print(f"\n\n===Logs for {request.node.name}===\n\n")
        for data_d in client.run_command("ls /var/log/sssd/").stdout_text.split():
            client.run_command(f'echo "--- {data_d} ---"; '
                               f'cat /var/log/sssd/{data_d}')


@pytest.fixture(scope="function")
def fetch_ca_cert(session_multihost, request, run_powershell_script):
    """ fixture to fetch CA Certificate and store in client """
    ret = run_powershell_script(name='getadcacert.ps1')
    if ret.returncode != 0:
        raise SSSDException("powershell script failed to execute")
    ca_crt = session_multihost.ad[0].get_file_contents('adca.der')
    session_multihost.client[0].put_file_contents('/tmp/adca.der', ca_crt)
    openssl_cmd = 'openssl x509 -inform der -in /tmp/adca.der '\
                  '-out /etc/openldap/certs/cacert.pem'
    cmd = session_multihost.client[0].run_command(openssl_cmd,
                                                  raiseonerr=False)
    if cmd.returncode != 0:
        pytest.fail("Failed to convert cert to ascii format")

    def remove_cert():
        """ Remove AD CA certificate """
        # remove /etc/openldap/certs/cacert.pem
        remove_client_ad_cacert = 'rm -f /etc/openldap/certs/cacert.pem'
        session_multihost.client[0].run_command(remove_client_ad_cacert)
    request.addfinalizer(remove_cert)


@pytest.fixture(scope="function")
def create_aduser_group(session_multihost, request):
    """ create AD user group """
    uid = random.randint(9999, 999999)
    ad = ADOperations(session_multihost.ad[0])
    ad_user = 'testuser%d' % uid
    ad_group = 'testgroup%d' % uid
    ad.create_ad_unix_user_group(ad_user, ad_group)

    def remove_ad_user_group():
        """ Remove windows AD user and group """
        ad.delete_ad_user_group(ad_group)
        ad.delete_ad_user_group(ad_user)

    request.addfinalizer(remove_ad_user_group)
    return (ad_user, ad_group)


@pytest.fixture(scope="function")
def create_nested_group(session_multihost, create_aduser_group, request):
    """ create nested AD group
        l1_grp is a top-level group
        l2_grp is a level-2 group and is a member of l1_grp group
        ad_user is a user-member of a 2nd-level l2_grp group
    """
    run_id = random.randint(999, 999999)
    ad_user = f'testuser-{run_id}'
    ad_group = f'testgroup-{run_id}'
    l1_grp = f'testgrp-l1-{run_id}'
    l2_grp = f'testgrp-l1-{run_id}'
    ad_op = ADOperations(session_multihost.ad[0])
    ad_op.create_ad_unix_group(l1_grp)
    ad_op.create_ad_unix_group(l2_grp)
    ad_op.create_ad_unix_user_group(ad_user, ad_group)
    ad_op.add_user_member_of_group(l1_grp, l2_grp)
    ad_op.add_user_member_of_group(l2_grp, ad_user)

    def remove_ad_user_group():
        """ Remove windows AD user and group """
        ad_op.delete_ad_user_group(ad_user)
        ad_op.delete_ad_user_group(ad_group)
        ad_op.delete_ad_user_group(l1_grp)
        ad_op.delete_ad_user_group(l2_grp)

    request.addfinalizer(remove_ad_user_group)
    return (run_id)


@pytest.fixture(scope="function")
def create_domain_local_group(session_multihost, request):
    """ Add user in domain local AD group"""
    ad_client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    for i in range(1, 6):
        ad_group = 'ltestgroup%d' % i
        ad_client.create_ad_unix_group(ad_group)

    def remove_ad_group():
        """ Remove AD user group """
        for i in range(1, 6):
            ad_group = 'ltestgroup%d' % i
            ad_client.remove_ad_user_group(ad_group)
    request.addfinalizer(remove_ad_group)


@pytest.fixture(scope="function")
def add_nisobject(session_multihost, request):
    """ Add nisobject to auto.direct map """
    share_list = [request.param]
    nfs_server = session_multihost.master[0].external_hostname
    client_ip = session_multihost.client[0].ip
    server = sssdTools(session_multihost.master[0])
    bkup = 'cp -af /etc/exports /etc/exports.backup'
    session_multihost.master[0].run_command(bkup)
    session_multihost.master[0].package_mgmt('nfs-utils', action='install')
    server.export_nfs_fs(share_list, client_ip)
    start_nfs = 'systemctl start nfs-server'
    try:
        session_multihost.master[0].run_command(start_nfs)
    except subprocess.CalledProcessError:
        pytest.fail("Unable to start nfs server")
    ad = ADOperations(session_multihost.ad[0])
    ret = ad.add_map(request.param, nfs_server)
    assert ret == 'Success'

    def remove_project():
        """ Remove the nisproject from map """
        stop_autofs = 'systemctl stop autofs'
        session_multihost.client[0].run_command(stop_autofs)
        ret = ad.delete_map(request.param)
        assert ret == 'Success'
        remove_share = 'rm -rf %s' % request.param
        session_multihost.master[0].run_command(remove_share)
        restore = 'cp -f /etc/exports.backup /etc/exports'
        session_multihost.master[0].run_command(restore)
        stop_nfs = "systemctl stop nfs-server"
        try:
            session_multihost.master[0].run_command(stop_nfs)
        except subprocess.CalledProcessError:
            pytest.fail("failed to stop nfs-server")
    request.addfinalizer(remove_project)


@pytest.fixture(scope="function")
def set_autofs_search_base(session_multihost, request):
    """ Enable autofs service """
    client = sssdTools(session_multihost.client[0])
    client.backup_sssd_conf()
    domain_name = client.get_domain_section_name()
    domain_section = 'domain/{}'.format(domain_name)
    domain_basedn_entry = session_multihost.ad[0].domain_basedn_entry
    autofs_dn = "ou=automount,%s" % (domain_basedn_entry)
    sssd_params = {'ldap_autofs_search_base': autofs_dn}
    client.sssd_conf(domain_section, sssd_params)

    def remove_autofs_search_base():
        """ Remove autofs search base """
        client.restore_sssd_conf()
        client.clear_sssd_cache()
    request.addfinalizer(remove_autofs_search_base)


@pytest.fixture(scope="function")
def add_user_in_domain_local_group(request, session_multihost,
                                   create_aduser_group):
    """ Add user in domain local AD group"""
    (ad_user, ad_group) = create_aduser_group

    ad_client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    for i in range(1, 6):
        ad_group = 'ltestgroup%d' % i
        ad_client.add_aduser_member_group(ad_group, ad_user)

    def remove_user_group():
        """ Remove Windows AD user and group """
        for i in range(1, 6):
            ad_group = 'ltestgroup%d' % i
            ad_client.remove_ad_user_group(ad_group)

    request.addfinalizer(remove_user_group)
    return ad_group


@pytest.fixture(scope="function")
def add_principals(session_multihost, request):
    """ Adds Service HTTPS and NFS principal to existing Host """
    client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    spn_list = ['HTTP', 'NFS']
    client.add_service_principals(spn_list)

    def remove_principals():
        """ Remove service principals """
        client.remove_service_principals(spn_list)
    request.addfinalizer(remove_principals)


@pytest.fixture(scope="function")
def adjoin(session_multihost, request):
    """ Join to AD using net ads command """
    ad_realm = session_multihost.ad[0].realm
    ad_ip = session_multihost.ad[0].ip
    ad_dc = session_multihost.ad[0].hostname
    client_ad = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    client_ad.disjoin_ad(raiseonerr=False)  # Make sure system is disjoined from AD
    client_ad.create_kdcinfo(ad_realm, ad_ip)
    kinit = "kinit Administrator"
    ad_password = session_multihost.ad[0].ssh_password
    try:
        session_multihost.client[0].run_command(kinit, stdin_text=ad_password)
    except subprocess.CalledProcessError:
        pytest.fail("kinit failed")

    def _join(membersw=None):
        """ Join AD """
        # We are using dc instead or realm to mitigate impact of AD being
        # on a different network than the client.
        # With "realm join REALM.COM" AD should be automatically discovered
        # if it is on the same network.
        # For all of the multi-arch tests we are mixing different networks
        # and it happens also sometimes with pure openstack.
        if membersw == 'samba':
            client_ad.join_ad(ad_dc, ad_password, mem_sw='samba')
        else:
            client_ad.join_ad(ad_dc, ad_password)
        session_multihost.client[0].run_command(
            f"cp -af {SSSD_DEFAULT_CONF} /etc/sssd/sssd.conf.adjoin")

    def adleave():
        """ Disjoin AD """
        session_multihost.client[0].run_command(
            f"cp -af /etc/sssd/sssd.conf.adjoin {SSSD_DEFAULT_CONF}")
        _adleave(client_ad)

    request.addfinalizer(adleave)
    return _join


@pytest.fixture(scope="function")
def get_rid(session_multihost, create_aduser_group):
    """
    Find Relative ID from object SID
    :param obj session_multihost: multihost object
    :Return: RID value
    """
    (user, _) = create_aduser_group
    client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    client.clear_sssd_cache()
    ad_user = '{}@{}'.format(user, session_multihost.ad[0].domainname)
    getent = 'getent passwd %s' % ad_user
    cmd = session_multihost.client[0].run_command(getent, raiseonerr=False)
    if cmd.returncode == 0:
        rid = client.find_rid(ad_user)
        return (ad_user, rid)
    else:
        pytest.fail("%s User lookup failed" % ad_user)


@pytest.fixture(scope="function")
def keytab_sssd_conf(session_multihost, request, adjoin):
    """ Add parameters required for keytab rotation in sssd.conf """
    adjoin(membersw='samba')
    client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    client.backup_sssd_conf()
    sssd_params = {'ad_maximum_machine_account_password_age': '1',
                   'ad_machine_account_password_renewal_opts': '300:15',
                   'debug_level': '9'}
    domain_name = client.get_domain_section_name()
    domain_section = 'domain/{}'.format(domain_name)
    client.sssd_conf(domain_section, sssd_params)
    client.clear_sssd_cache()

    def restore_sssd_conf():
        """ Restore original sssd.conf """
        client.restore_sssd_conf()
    request.addfinalizer(restore_sssd_conf)


@pytest.fixture(scope="function")
def cifsmount(session_multihost, request):
    """ Mount cifs share and create files with
    different permissions
    """
    ad_user = 'idmfoouser1'
    kinit = 'kinit %s' % ad_user
    server = session_multihost.master[0].sys_hostname.strip().split('.')[0]
    session_multihost.client[0].run_command(kinit, stdin_text='Secret123')
    mountcifs = "mount -t cifs -o cifsacl "\
                "-o sec=krb5 -o username=%s //%s/share1"\
                " /mnt/samba/share1" % (ad_user, server)
    cmd = session_multihost.client[0].run_command(mountcifs, raiseonerr=False)
    time.sleep(5)
    if cmd.returncode != 0:
        journalctl = 'journalctl -x -n 50 --no-pager'
        session_multihost.client[0].run_command(journalctl)

    def cifsunmount():
        """ Umount the cifs shares """
        umount = "umount /mnt/samba/share1"
        cmd = session_multihost.client[0].run_command(umount, raiseonerr=False)
        assert cmd.returncode == 0
        kdestroy = 'kdestroy -A'
        session_multihost.client[0].run_command(kdestroy, raiseonerr=False)
    request.addfinalizer(cifsunmount)


@pytest.fixture(scope='function')
def backupsssdconf(session_multihost, request):
    """ Backup and restore sssd.conf """
    tools = sssdTools(session_multihost.client[0])
    tools.backup_sssd_conf()
    session_multihost.client[0].service_sssd('stop')

    def restoresssdconf():
        """ Restore sssd.conf """
        tools.restore_sssd_conf()
    request.addfinalizer(restoresssdconf)


@pytest.fixture(scope="function")
def range_retr_mods(session_multihost, fetch_ca_cert, request):
    """ AD default query policy modifications, MaxValRange is changed
        from 1500 to 50"""
    basedn = session_multihost.ad[0].domain_basedn_entry
    DEF_QUERY = 'cn=Default Query Policy,cn=Query-Policies,cn=Directory Service,'\
        f'cn=Windows NT,cn=Services,cn=Configuration,{basedn}'
    with tempfile.NamedTemporaryFile(mode='w', newline='\n') as tfile:
        tfile.write(f"dn: {DEF_QUERY}\n")
        tfile.write("changetype: modify\n")
        tfile.write("delete: lDAPAdminLimits\n")
        tfile.write("lDAPAdminLimits: MaxValRange=1500\n")
        tfile.write("-\n")
        tfile.write("add: lDAPAdminLimits\n")
        tfile.write("lDAPAdminLimits: MaxValRange=50\n")
        tfile.write("-\n")
        tfile.flush()
        session_multihost.ad[0].transport.put_file(
            tfile.name, '/home/Administrator/mod_maxrange.ldif')
    session_multihost.ad[0].run_command('ldifde.exe -i -f mod_maxrange.ldif', raiseonerr=False)


# ############## class scoped Fixtures ##############################


@pytest.fixture(scope="class")
def multihost(session_multihost, request):
    """ Multihost fixture to be used by tests
    :param obj session_multihost: multihost object
    :return obj session_multihost: return multihost object
    :Exceptions: None
    """
    if hasattr(request.cls(), 'class_setup'):
        request.cls().class_setup(session_multihost)
        request.addfinalizer(
            lambda: request.cls().class_teardown(session_multihost))
    return session_multihost


@pytest.fixture(scope="class")
def clear_sssd_cache(session_multihost):
    """ Clear sssd cache """
    client = sssdTools(session_multihost.client[0])
    client.clear_sssd_cache()


@pytest.fixture(scope="class")
def enable_autofs_schema(session_multihost, request):
    """ Enable autofs schema(rfc2307) to Windows AD.
    :param obj session_multihost: multihost object
    :param obj request: pytest request object
    """
    ad_realm = session_multihost.ad[0].realm
    ad_password = session_multihost.ad[0].ssh_password
    ad_client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    try:
        ad_client.remove_automount(verbose=False)
    except subprocess.CalledProcessError:
        print("Automount entry not found")
    else:
        print("Existing automount entry deleted")
    realm_output = ad_client.join_ad(ad_realm, ad_password)
    ad = ADOperations(session_multihost.ad[0])
    ad.add_autofs_schema()

    def remove_automount_entries():
        """ Remove autofs schema from Windows AD """
        ad_client.disjoin_ad(realm_output)
        ad_client.remove_automount()
    request.addfinalizer(remove_automount_entries)


@pytest.fixture(scope="class")
def enable_autofs_service(session_multihost,
                          request):  # pylint: disable=unused-argument
    """ Enable autofs service """
    client = sssdTools(session_multihost.client[0])
    client.backup_sssd_conf()
    services = {'services': 'nss, pam, autofs'}
    client.sssd_conf('sssd', services)
    domain_name = client.get_domain_section_name()
    domain_section = 'domain/{}'.format(domain_name)
    sssd_params = {'autofs_provider': 'ad',
                   'debug_level': '9'}
    client.sssd_conf(domain_section, sssd_params)

    def disable_autofs_service():
        """ Disable autofs service in sssd.conf """
        services = {'services': 'nss, pam'}
        client.sssd_conf('sssd', services)
        client.clear_sssd_cache()
        stop_autofs = 'systemctl stop autofs'
        session_multihost.client[0].run_command(stop_autofs)
    request.addfinalizer(disable_autofs_service)


@pytest.fixture(scope="class")
def create_idmap_users_groups(session_multihost, request):
    """ create Ad idmap users and groups """
    ad = ADOperations(session_multihost.ad[0])
    usrgrp = ['posix_usr', 'posix_grp', 'noposix_usr', 'noposix_grp']
    for object in usrgrp:
        ad.delete_ad_user_group(object)
    ad.create_ad_nonposix_user('noposix_usr')
    ad.create_ad_nonposix_group('noposix_grp')
    ad.add_user_member_of_group('noposix_grp', 'noposix_usr')
    ad.create_ad_unix_user_group('posix_usr', 'posix_grp')

    def remove_idmap_objects():
        """ Remove AD idmap users and groups """
        for object in usrgrp:
            ad.delete_ad_user_group(object)
    request.addfinalizer(remove_idmap_objects)


@pytest.fixture(scope="class")
def create_small_grp_usr(session_multihost, request):
    """ Create a group with 50 member users """
    ad = ADOperations(session_multihost.ad[0])
    AD_MAX_RANGE = 51
    ad.create_ad_unix_group('smallgrp')
    for uid in range(1, AD_MAX_RANGE):
        ad.create_ad_unix_user(f'ad_user{uid}')
        ad.add_user_member_of_group('smallgrp', f'ad_user{uid}')

    def remove_ad_user_group():
        """ Remove windows AD user and group """
        ad.delete_ad_user_group('smallgrp')
        for uid in range(1, AD_MAX_RANGE):
            ad.delete_ad_user_group(f'ad_user{uid}')
    request.addfinalizer(remove_ad_user_group)


@pytest.fixture(scope="class")
def create_range_aduser_group(session_multihost, request):
    """ create 200 AD users and groups
        Creates a group with 200 member users
        Creates a user with the 200 groups membership
    """
    ad = ADOperations(session_multihost.ad[0])
    ad.create_ad_unix_user('rangeuser')
    ad.create_ad_unix_group('rangegroup')
    for uid in range(1, 201):
        ad.create_ad_unix_user(f'rangeuser0{uid}')
        ad.create_ad_unix_group(f'rangegroup0{uid}')
        ad.add_user_member_of_group('rangegroup', f'rangeuser0{uid}')
        ad.add_user_member_of_group(f'rangegroup0{uid}', 'rangeuser')

    def remove_ad_user_group():
        """ Remove windows AD user and group """
        for uid in range(1, 201):
            ad_user = f'rangeuser0{uid}'
            ad_group = f'testgroup0{uid}'
            ad.delete_ad_user_group(ad_group)
            ad.delete_ad_user_group(ad_user)
    request.addfinalizer(remove_ad_user_group)


@pytest.fixture(scope="class")
def enable_ad_sudoschema(session_multihost):
    """ Enable AD Sudo schema """
    basedn = session_multihost.ad[0].domain_basedn_entry
    ldapuri = 'ldap://%s' % (session_multihost.ad[0].ip)
    print(session_multihost.ad[0].ip)
    password = session_multihost.ad[0].ssh_password
    user = 'cn=Administrator,cn=Users,%s' % (basedn)
    searchbase = 'CN=sudoRole,CN=Schema,CN=Configuration,%s' % basedn
    ldapsearch = 'ldapsearch -x -LLL -b %s -D %s -w %s'\
                 ' -H %s cn=SudoRole' % (searchbase, user, password, ldapuri)
    cmd = session_multihost.client[0].run_command(ldapsearch, raiseonerr=False)
    if cmd.returncode == 0:
        print("Schema already added")
    else:
        cwd = os.path.dirname(os.path.abspath(__file__))
        split_cwd = cwd.split('/')
        idx = split_cwd.index('multihost')
        path_list = split_cwd[:idx + 1]
        sssd_qe_path = '/'.join(path_list)
        data_path = "%s/data" % sssd_qe_path
        filename = 'schema.ActiveDirectory'
        remote_file_path = posixpath.join('/home/administrator', filename)
        source_file_path = posixpath.join(data_path, filename)
        session_multihost.ad[0].transport.put_file(source_file_path,
                                                   remote_file_path)
        ldifde = 'ldifde.exe -i -f schema.ActiveDirectory'\
                 ' -c dc=X %s' % basedn
        session_multihost.ad[0].run_command(ldifde, raiseonerr=False)


@pytest.fixture(scope="class")
def create_ad_sudousers(session_multihost, request):
    """ create Ad sudo users and groups """
    ad = ADOperations(session_multihost.ad[0])
    for idx in range(1, 10):
        ad_user = 'sudo_idmuser%d' % idx
        ad_group = 'sudo_idmgroup%d' % idx
        ad.delete_ad_user_group(ad_group)
        ad.delete_ad_user_group(ad_user)
        ad.create_ad_unix_user_group(ad_user, ad_group)
    ad_user1 = 'sudo_user1'
    ad_group2 = 'sudo_group2'
    ad.create_ad_unix_user_group(ad_user1, ad_group2)
    ad_group1 = 'sudo_groupx'
    ad.create_ad_nonposix_group(ad_group1)
    ad.add_user_member_of_group(ad_group1, ad_user1)
    ad_user_a = 'sudo_usera'
    ad_group_a = 'sudo_groupa'
    ad.create_ad_unix_user_group(ad_user_a, ad_group_a)

    def remove_ad_sudousers():
        """ Remove AD sudo users and groups """
        for idx in range(1, 10):
            ad_user = 'sudo_idmuser%d' % idx
            ad_group = 'sudo_idmgroup%d' % idx
            ad.delete_ad_user_group(ad_group)
            ad.delete_ad_user_group(ad_user)
        usrgrp = [ad_user1, ad_group1, ad_group2, ad_user_a, ad_group_a]
        for object in usrgrp:
            ad.delete_ad_user_group(object)
    request.addfinalizer(remove_ad_sudousers)


@pytest.fixture(scope="class")
def sudorules(session_multihost, request):
    """ Create AD Sudo rules """
    basedn = session_multihost.ad[0].domain_basedn_entry
    realm = session_multihost.ad[0].realm
    winad = ADOperations(session_multihost.ad[0])
    sudo_ou = f'ou=Sudoers,{basedn}'
    sudo_options = ["!requiretty", "!authenticate"]

    # Delete and recreate sudo OU to make sure that there are no leftovers
    winad.del_sudo_ou(verbose=False, raiseonerr=False)
    winad.add_sudo_ou()

    for item in ['user', 'group']:
        for idx in range(1, 10):
            rule_dn = f'cn=less_{item}_rule{idx},{sudo_ou}'
            sudo_identity = [f'sudo_idm{item}{idx}@{realm}']
            if item == 'user':
                sudo_identity.append(f'sudo_idmuser{idx}')
            res = winad.add_sudo_rule(
                rule_dn, 'ALL', '/usr/bin/less', sudo_identity, sudo_options)
            if not res:
                pytest.fail(f"Failed to add sudo rule {rule_dn}")

    rule_dn = f'cn=%%head_nonposix_rule,{sudo_ou}'
    sudo_identity = [f'%%sudo_groupx@{realm}', '%sudo_groupx']
    res = winad.add_sudo_rule(rule_dn, 'ALL', '/usr/bin/head',
                              sudo_identity, sudo_options)
    if not res:
        pytest.fail(f"Failed to add sudo rule {rule_dn}")

    rule_dn = f'cn=head_rule1,{sudo_ou}'
    sudo_options = ["!requiretty", "!authenticate"]
    res = winad.add_sudo_rule(rule_dn, 'ALL', '/usr/bin/head', 'sudo_usera',
                              sudo_options, runas='sudo_idmuser1')
    if not res:
        pytest.fail(f"Failed to add sudo rule {rule_dn}")

    def delete_ad_sudorules():
        """ Delete sudo OU with all of the sudo rules."""
        winad.del_sudo_ou(verbose=False, raiseonerr=False)
    request.addfinalizer(delete_ad_sudorules)


@pytest.fixture(scope="class")
def joinad(session_multihost, request):
    """ class fixture to join AD using realm """
    client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    client.disjoin_ad(raiseonerr=False)  # Make sure system is disjoined from AD
    kinit = "kinit Administrator"
    ad_password = session_multihost.ad[0].ssh_password
    ad_dc = session_multihost.ad[0].hostname
    client.join_ad(realm=ad_dc)
    try:
        session_multihost.client[0].service_sssd('restart')
    except SSSDException:
        cmd = f'cat {SSSD_DEFAULT_CONF}'
        session_multihost.client[0].run_command(cmd)
        journal = 'journalctl -x -n 150 --no-pager'
        session_multihost.client[0].run_command(journal)
    retry = 0
    while (retry != 5):
        cmd = session_multihost.client[0].run_command(kinit,
                                                      stdin_text=ad_password,
                                                      raiseonerr=False)
        if cmd.returncode == 0:
            break
        else:
            retry += 1
            time.sleep(5)

    def disjoin():
        """ Disjoin system from Windows AD """
        _adleave(client)
    request.addfinalizer(disjoin)


@pytest.fixture(scope="class")
def winbind_server(session_multihost, request):
    """ Winbind Server """
    master = sssdTools(session_multihost.master[0], session_multihost.ad[0])
    client = sssdTools(session_multihost.client[0], session_multihost.ad[0])
    master.server_install_pkgs()
    smb_master = sambaTools(session_multihost.master[0],
                            session_multihost.ad[0])
    smb_master.enable_winbind()
    smb_client = sambaTools(session_multihost.client[0],
                            session_multihost.ad[0])
    smb_client.enable_winbind()

    def disable():
        """ Disable winbind """
        print("we are disabling winbind")
        master.disjoin_ad()
        client.disjoin_ad()
        smb_master.disable_winbind()
        smb_master.clear_samba_cache()
        smb_master.remove_smbconf()
        smb_client.disable_winbind()
        smb_client.clear_samba_cache()
        smb_client.remove_smbconf()
    request.addfinalizer(disable)


@pytest.fixture(scope='class')
def configure_samba(session_multihost, request):
    """ samba server """
    master = sambaTools(session_multihost.master[0], session_multihost.ad[0])
    master.add_share_definition('share1', '/mnt/samba/share1')
    master.service_smb(action='restart')
    time.sleep(20)

    def stop_samba_server():
        """ Stop samba server """
        print("we are stopping samba server")
        master.service_smb(action='stop')
        master.clear_samba_cache()
        master.remove_smbconf()
    request.addfinalizer(stop_samba_server)


@pytest.fixture(scope='class')
def samba_share_permissions(session_multihost, request):
    """ Set permissions on samba share """
    smbTools = sambaTools(session_multihost.master[0], session_multihost.ad[0])
    adops = ADOperations(session_multihost.ad[0])
    share_name = 'share1'
    share_path = '/mnt/samba/%s' % share_name
    smbTools.create_samba_share(share_path)
    realm = session_multihost.ad[0].realm
    for idx in range(1, 3):
        ad_user = 'idmfoouser%d' % idx
        ad_group = 'idmfoogroup%d' % idx
        all_group = 'idmfooallgroup'
        adops.delete_ad_user_group(ad_group)
        adops.delete_ad_user_group(ad_user)
    adops.delete_ad_user_group(all_group)
    adops.create_ad_unix_group(all_group)
    for idx in range(1, 3):
        ad_user = 'idmfoouser%d' % idx
        ad_group = 'idmfoogroup%d' % idx
        adops.create_ad_unix_user_group(ad_user, ad_group)
        adops.add_user_member_of_group(all_group, ad_user)
    session_multihost.master[0].service_sssd('restart')
    time.sleep(30)

    for idx in range(1, 3):
        ad_user = 'idmfoouser%d' % idx
        ad_group = 'idmfoogroup%d' % idx
        directory = '/mnt/samba/share1/idmfoogroup%d' % idx
        create_dir = 'mkdir -p %s' % directory
        session_multihost.master[0].run_command(create_dir)
        chmod = 'chmod 2770 %s' % directory
        session_multihost.master[0].run_command(chmod)
        chgrp = "chgrp '%s@%s' %s " % (ad_group, realm, directory)
        session_multihost.master[0].run_command(chgrp)

    all_group = 'idmfooallgroup'
    common_dir = 'mkdir -p /mnt/samba/share1/allgroup'
    session_multihost.master[0].run_command(common_dir)
    chgrp = "chgrp '%s@%s' /mnt/samba/share1/allgroup " % (all_group, realm)
    chmod = "chmod 2770 /mnt/samba/share1/allgroup"
    session_multihost.master[0].run_command(chgrp)
    session_multihost.master[0].run_command(chmod)

    # create mount point on client
    mount_point = 'mkdir -p %s' % share_path
    session_multihost.client[0].run_command(mount_point)

    def delete_share_directory():
        """ Delete share directory """
        print("we are deleting samba share directory")
        smbTools.delete_samba_share(share_path)
        remove_mount_point = "rm -rf %s" % share_path
        session_multihost.client[0].run_command(remove_mount_point)
        for idx in range(1, 3):
            ad_user = 'idmfoouser%d' % idx
            ad_group = 'idmfoogroup%d' % idx
            all_group = 'idmfooallgroup'
            adops.delete_ad_user_group(ad_group)
            adops.delete_ad_user_group(all_group)
            adops.delete_ad_user_group(ad_user)
    request.addfinalizer(delete_share_directory)

# ################### Session scoped fixtures #########################


@pytest.fixture(scope='session', autouse=True)
def fips_ad_support_policy(session_multihost, request):
    """ Enable FIPS:AD-SUPPORT crypto policy added in bz2056676"""
    old_policy = session_multihost.client[0].run_command(
        'update-crypto-policies --show', raiseonerr=False).stdout_text
    old_policy = old_policy.strip()
    if "FIPS" == old_policy:
        session_multihost.client[0].run_command(
            'update-crypto-policies --set FIPS:AD-SUPPORT', raiseonerr=False)
    old_policy_master = session_multihost.master[0].run_command(
        'update-crypto-policies --show', raiseonerr=False).stdout_text
    old_policy_master = old_policy_master.strip()
    if "FIPS" == old_policy_master:
        session_multihost.master[0].run_command(
            'update-crypto-policies --set FIPS:AD-SUPPORT', raiseonerr=False)

    def restore_policy():
        """ Restore crypto policy """
        if "FIPS" == old_policy:
            session_multihost.client[0].run_command(
                f'update-crypto-policies --set {old_policy}', raiseonerr=False)
        if "FIPS" == old_policy_master:
            session_multihost.master[0].run_command(
                f'update-crypto-policies --set {old_policy_master}',
                raiseonerr=False
            )
    request.addfinalizer(restore_policy)


@pytest.fixture(scope="session", autouse=True)
def setup_session(request, session_multihost, create_testdir):
    """ Setup Session """
    client = sssdTools(session_multihost.client[0])
    realm = session_multihost.ad[0].realm
    ad_host = session_multihost.ad[0].sys_hostname
    try:
        master = sssdTools(session_multihost.master[0])
    except IndexError:
        pass
    else:
        master.server_install_pkgs()
        master.update_resolv_conf(session_multihost.ad[0].ip)
    client.client_install_pkgs()
    client.update_resolv_conf(session_multihost.ad[0].ip)
    client.systemsssdauth(realm, ad_host)

    def teardown_session():
        """ Teardown session """
        session_multihost.client[0].service_sssd('stop')
        remove_sssd_conf = f'rm -f {SSSD_DEFAULT_CONF}'
        session_multihost.client[0].run_command(remove_sssd_conf)
    request.addfinalizer(teardown_session)



"""Tests related to host map and network map

:requirement: SSSD NSS should support host map
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
from sssd.testlib.common.utils import sssdTools, LdapOperations
import pytest


@pytest.mark.usefixtures('setup_sssd', 'enable_sssd_hostmap', 'add_host_entry')
@pytest.mark.hostmap
class TestHostMaps(object):
    """
    SSSD NSS should support host map
    :bugzilla:
     https://bugzilla.redhat.com/show_bug.cgi?id=1340908
    """
    @pytest.mark.tier1
    def test_001_ldap_iphost_search_base(self, multihost):
        """
        :title: hosts: Test ldap_iphost_search_base field
        :id: a5b82f87-e4f2-49ce-bfc2-e23f5020219d
        :customerscenario: True
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_iphost_search_base':
                         'ou=People,dc=example,dc=test'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent = "getent hosts node1"
        cmd = multihost.client[0].run_command(getent)
        if '192.168.1.1' in cmd.stdout_text and \
                '192.168.1.2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        assert status == 'PASS'
        getent = "getent hosts 192.168.1.1"
        cmd = multihost.client[0].run_command(getent)
        if 'node1' in cmd.stdout_text and 'node2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_002_ldap_iphost_object_class(self, multihost):
        """
        :title: hosts: Test ldap_iphost_object_class field
        :id: eed161fa-1310-43b7-90e6-dd99846d9cbe
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_iphost_object_class': 'ipHost'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent = "getent hosts node1"
        cmd = multihost.client[0].run_command(getent)
        if '192.168.1.1' in cmd.stdout_text and \
                '192.168.1.2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        assert status == 'PASS'
        getent = "getent hosts 192.168.1.1"
        cmd = multihost.client[0].run_command(getent)
        if 'node1' in cmd.stdout_text and 'node2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_003_ldap_iphost_name(self, multihost):
        """
        :title: hosts: Test ldap_iphost_name field
        :id: 00ae3874-f3ef-4121-bb89-1fce554ddbef
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_iphost_name': 'cn'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent = "getent hosts node1"
        cmd = multihost.client[0].run_command(getent)
        if '192.168.1.1' in cmd.stdout_text and \
                '192.168.1.2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        assert status == 'PASS'
        getent = "getent hosts 192.168.1.1"
        cmd = multihost.client[0].run_command(getent)
        if 'node1' in cmd.stdout_text and 'node2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_004_ldap_iphost_number(self, multihost):
        """
        :title: hosts: Test ldap_iphost_number field
        :id: f78b0874-411c-45f4-8497-e1544d01fdbc
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_iphost_number': 'ipHostNumber'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent = "getent hosts node1"
        cmd = multihost.client[0].run_command(getent)
        if '192.168.1.1' in cmd.stdout_text and \
                '192.168.1.2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        assert status == 'PASS'
        getent = "getent hosts 192.168.1.1"
        cmd = multihost.client[0].run_command(getent)
        if 'node1' in cmd.stdout_text and 'node2' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_005_ldap_ipnetwork_search_base(self, multihost):
        """
        :title: hosts: Test ldap_ipnetwork_search_base field
        :id: 075bd1e0-8db2-4cc4-bfdf-99fbbe8df62d
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_ipnetwork_search_base':
                         'ou=People,dc=example,dc=test'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent_networks = "getent networks node2"
        cmd = multihost.client[0].run_command(getent_networks)
        if '192.168.1.2' in cmd.stdout_text and \
                '192.168.1.1' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_006_ldap_ipnetwork_object_class(self, multihost):
        """
        :title: hosts: Test ldap_ipnetwork_object_class field
        :id: fcf249c4-4501-4ff3-a916-5ab13cccc349
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_ipnetwork_object_class': 'ipNetwork'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent_networks = "getent networks node2"
        cmd = multihost.client[0].run_command(getent_networks)
        if '192.168.1.2' in cmd.stdout_text and \
                '192.168.1.1' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_007_ldap_ipnetwork_name(self, multihost):
        """
        :title: hosts: Test ldap_ipnetwork_name field
        :id: 8d90a533-7d42-4dd4-aab1-c8c3e25cb60a
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_ipnetwork_name': 'cn'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent_networks = "getent networks node2"
        cmd = multihost.client[0].run_command(getent_networks)
        if '192.168.1.2' in cmd.stdout_text and \
                '192.168.1.1' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_008_ldap_ipnetwork_number(self, multihost):
        """
        :title: hosts: Test ldap_ipnetwork_number field
        :id: ae1df923-5115-4f58-8650-d4a187cf8f7e
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_ipnetwork_number': 'ipNetworkNumber'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent_networks = "getent networks node2"
        cmd = multihost.client[0].run_command(getent_networks)
        if '192.168.1.2' in cmd.stdout_text and \
                '192.168.1.1' not in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_009_ipnetwork_iphost(self, multihost):
        """
        :title: hosts: Test ldap_ipnetwork_search_base
         and ldap_iphost_search_base fields
        :id: 6394baf6-d285-40ef-8182-262ef29d4e33
        """
        client = sssdTools(multihost.client[0])
        domain_params = {
            'ldap_ipnetwork_search_base': 'ou=People,dc=example,dc=test',
            'ldap_iphost_search_base': 'ou=People,dc=example,dc=test'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        getent_networks = "getent networks node3"
        cmd = multihost.client[0].run_command(getent_networks)
        if '192.168.1.3' in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        assert status == 'PASS'
        getent_networks = "getent hosts 192.168.1.3"
        cmd = multihost.client[0].run_command(getent_networks)
        if 'node3' in cmd.stdout_text:
            status = 'PASS'
        else:
            status = 'FAIL'
        client.sssd_conf('domain/example1', domain_params, action='delete')
        assert status == 'PASS'

    @pytest.mark.tier1_3
    def test_more_than_one_cn(self, multihost):
        """
        :title: hosts: 'getent hosts' not return hosts
         if they have more than one CN in LDAP
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2056035
        :id: f2bacc70-ccff-11ec-9b25-845cf3eff344
        :customerscenario: true
        :steps:
          1. Create hostname having multiple 'cn'
          2. Entry having multiple 'cn' handled correctly by SSSD
        :expectedresults:
          1. Should succeed
          2. Should succeed
        """
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_info = {'cn': ['node4'.encode('utf-8'),
                            'node4.example.test'.encode('utf-8')],
                     'objectClass': [b'top', b'ipHost', b'device'],
                     'ipHostNumber': '192.168.1.4'.encode('utf-8')}
        user_dn = 'cn=node4+ipHostNumber=192.168.1.4,ou=People,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_iphost_search_base': 'ou=People,dc=example,dc=test'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        output = multihost.client[0].run_command("getent hosts node4").stdout_text
        for i in ['node4.example.test', '192.168.1.4']:
            assert i in output
        output = multihost.client[0].run_command("getent "
                                                 "hosts "
                                                 "node4.example.test").stdout_text
        for i in ['node4.example.test', '192.168.1.4']:
            assert i in output
        cmd = multihost.client[0].run_command("getent "
                                              "hosts "
                                              "node1 "
                                              "node3 "
                                              "node4").stdout_text
        for i in ['192.168.1.1', '192.168.1.3', '192.168.1.4',
                  'node1', 'node3', 'node4.example.test']:
            assert i in cmd
        ldap_inst.del_dn('cn=node4+ipHostNumber=192.168.1.4,ou=People,dc=example,dc=test')



"""
SSSD AD Provider Test Cases

:requirement: ad
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericADProvider
from sssd_test_framework.topology import KnownTopologyGroup


@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(jira="RHEL-65848", gh=7690)
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.importance("high")
def test_ad__user_authentication_when_provider_is_set_to_ldap_with_gss_spnego(
    client: Client, provider: GenericADProvider, method: str
):
    """
    :title: Login to AD when id_provider is set to ldap
    :setup:
        1. Add AD user
        2. Update sssd.conf with 'id_provider = ldap', 'ldap_schema = ad',
        'ldap_id_use_start_tls = false', 'auth_provider = ad' and
        'ldap_sasl_mech = gssspengo' and Start SSSD
    :steps:
        1. Check authentication of the user
        2. Check log message in krb5_child.log, UPN [user1@null] should not be logged
    :expectedresults:
        1. Authentication is successful
        2. Get required UPN [user1@<domain_name>] from krb5_child.log
    :customerscenario: False
    """
    provider.user("user1").add()

    client.sssd.config.remove_option("domain/test", "id_provider")

    configurations = {
        "id_provider": "ldap",
        "ldap_schema": "ad",
        "ldap_id_use_start_tls": "False",
        "auth_provider": "ad",
        "ldap_referrals": "False",
        "ldap_sasl_mech": "GSS-SPNEGO",
        "ldap_id_mapping": "True",
    }

    for key, value in configurations.items():
        client.sssd.domain[key] = value

    # id_provider = ldap will not add them automatically if they are not
    # defined on the server side.
    client.sssd.nss["default_shell"] = "/bin/bash"
    client.sssd.nss["override_homedir"] = "/home/%u"

    # `provider.host.domain` is ignored because it is dynamically added
    p_domain = f"{provider.host.domain}"  # type: ignore[attr-defined]

    client.sssd.domain["krb5_realm"] = f"{p_domain.upper()}"
    client.sssd.domain["dns_discovery_domain"] = f"{p_domain}"

    client.sssd.start()

    assert client.auth.parametrize(method).password("user1", "Secret123"), "User user1 failed login!"

    log_str = client.fs.read("/var/log/sssd/krb5_child.log")
    assert f"UPN [user1@{p_domain}]" in log_str, f"'UPN [user1@{p_domain}]' not in logs!"  # type: ignore



"""
SSSD Cache Tests.

Tests pertaining SSSD caches, the following types are tested and some will be in other python files.

* Local cache (LDB)
* Negative cache (ncache)
* In-memory cache (memcache): test_memcache.py

Note: There is not added benefit to test against all topologies, the cache tests are tested against LDAP.

:requirement: Cache
"""

from __future__ import annotations

import time

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__entries_are_refreshed_as_configured(client: Client, provider: LDAP):
    """
    :title: Ensuring LDB cache refreshes at configured intervals
    :setup:
        1. Create user
        2. Create group
        3. Create netgroup
        4. Configure SSSD and set 'entry_cache_timeout to 1' and 'refresh_expired_interval to 2'
        5. Restart SSSD
        6. Lookup user, group and netgroup
    :steps:
        1. Search for objects lastUpdate and dataExpireTimestamp in ldb database
        2. Wait 5 seconds and repeat search
    :expectedresults:
        1. The 'dataExpireTimestamp' value equals the 'lastUpdate + entry_cache_timeout' value
        2. Objects 'lastUpdate' timestamp value has been refreshed
    :customerscenario: False
    """
    user = provider.user("test_user").add()
    provider.group("test_group").add().add_member(user)
    provider.netgroup("test_netgroup").add().add_member(user=user)

    domain = client.sssd.default_domain
    entry_cache_timeout = 1
    refresh_expired_interval = 2

    client.sssd.domain["entry_cache_timeout"] = str(entry_cache_timeout)
    client.sssd.domain["refresh_expired_interval"] = str(refresh_expired_interval)

    client.sssd.restart()
    client.tools.getent.passwd(f"test_user@{domain}")
    client.tools.getent.group(f"test_group@{domain}")
    client.tools.getent.netgroup(f"test_netgroup@{domain}")

    ldb_cache = f"/var/lib/sss/db/cache_{domain}.ldb"
    ldb_suffix = f"cn={domain},cn=sysdb"

    last_update: list[int] = []
    expire_time: list[int] = []

    for i in [f"test_user@{domain}", f"test_group@{domain}", "test_netgroup"]:
        result = client.ldb.search(ldb_cache, ldb_suffix, filter=f"name={i}")
        for k, v in result.items():
            for y in v.items():
                if y[0] == "lastUpdate":
                    last_update = last_update + [(int(y[1][0]))]
                if y[0] == "dataExpireTimestamp":
                    expire_time = expire_time + [(int(y[1][0]))]

    for m, n in enumerate(last_update):
        assert (
            last_update[m] + entry_cache_timeout == expire_time[m]
        ), f"{expire_time[m]} != {last_update[m]} + {entry_cache_timeout}"

    time.sleep(5)

    for s, t in enumerate([f"test_user@{domain}", f"test_group@{domain}", "test_netgroup"]):
        result = client.ldb.search(ldb_cache, ldb_suffix, filter=f"name={t}")
        for k, v in result.items():
            for y in v.items():
                if y[0] == "lastUpdate":
                    assert last_update[s] <= (int(y[1][0])), f"{s} lastUpdate value is greater than expected!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__writes_to_both_database_files(client: Client, provider: LDAP):
    """
    :title: Search for user in the following ldb databases, cache_*.ldb and timestamp_*.ldb
    :setup:
        1. Create user
        2. Start SSSD
    :steps:
        1. Lookup user
        2. Check cache
        3. Lookup user in cache ldb database
        4. Lookup user in timestamp ldb database
    :expectedresults:
        1. User is found
        2. Cache file exists
        3. User found
        4. User found
    :customerscenario: False
    """
    provider.user("user1").add()
    client.sssd.start()
    client.tools.getent.passwd("user1")
    cache = "/var/lib/sss/db/cache_test.ldb"
    timestamps = "/var/lib/sss/db/timestamps_test.ldb"
    assert client.fs.exists(timestamps), f"Timestamp file '{timestamps}' does not exist"

    ldb1 = client.ldb.search(cache, f"name=user1@{client.sssd.default_domain},cn=users,cn=test,cn=sysdb")
    ldb2 = client.ldb.search(timestamps, f"name=user1@{client.sssd.default_domain},cn=users,cn=test,cn=sysdb")

    assert ldb1, f"ldbsearch failed to find user1 in {cache}!"
    assert ldb2, f"ldbsearch failed to find user1 in {timestamps}!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__writes_to_both_database_files_when_using_fully_qualified_names(client: Client, provider: LDAP):
    """
    :title: Search for user using fully qualified name in the following ldb databases, cache_*.ldb and timestamp_*.ldb
    :setup:
        1. Create user
        2. Start SSSD
    :steps:
        1. Lookup user
        2. Lookup user in cache ldb database
        3. Lookup user in timestamp ldb database
    :expectedresults:
        1. User found
        2. User found
        3. User found
    :customerscenario: False
    """
    provider.user("user1").add()
    client.sssd.domain["use_fully_qualified_names"] = "True"
    client.sssd.start()
    client.tools.getent.passwd(f"user1@{client.sssd.default_domain}")

    cache = "/var/lib/sss/db/cache_test.ldb"
    timestamps = "/var/lib/sss/db/timestamps_test.ldb"
    user_dn = f"name=user1@{client.sssd.default_domain},cn=users,cn=test,cn=sysdb"
    ldb1 = client.ldb.search(cache, user_dn)
    ldb2 = client.ldb.search(timestamps, user_dn)

    assert ldb1, f"ldbsearch failed to find user1 in {cache}!"
    assert ldb2, f"ldbsearch failed to find user1 in {timestamps}!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__user_entries_contains_latest_changes_when_modified_and_deleted(client: Client, provider: LDAP):
    """
    :title: Checks user changes are reflected when modified and deleted
    :setup:
        1. Add users 'user1' and 'user2'
        2. Start SSSD
        3. Lookup users
    :steps:
        1. Login as users
        2. Modify 'user1' shell and delete 'user2' and clear cache
        3. Login as users
        4. Lookup user 'user2'
        5. Lookup user 'user1'
    :expectedresults:
        1. Users logged in
        2. User 'user1' is modified and user 'user2' is deleted
        3. User 'user1' logged in
        4. User 'user2' is not found
        5. User 'user1' is found and shell was updated
    :customerscenario: False
    """
    provider.user("user1").add(shell="/bin/bash")
    provider.user("user2").add(shell="/bin/bash")
    client.sssd.start()
    client.tools.getent.passwd("user1")
    client.tools.getent.passwd("user2")

    assert client.auth.ssh.password("user1", "Secret123"), "Login failed!"
    assert client.auth.ssh.password("user2", "Secret123"), "Login failed!"

    provider.user("user2").delete()
    provider.user("user1").modify(shell="/bin/sh")

    client.sssctl.cache_expire(everything=True)

    assert client.auth.ssh.password("user1", "Secret123"), "Login failed!"
    assert not client.auth.ssh.password("user2", "Secret123"), "Login successful!"

    result = client.tools.getent.passwd("user1")
    assert result is not None, "User not found!"
    assert result.shell == "/bin/sh", "User shell did not update!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__group_entries_contains_latest_changes_when_modified_and_deleted(
    client: Client,
    provider: LDAP,
):
    """
    :title: Check latest group changes are reflected when modified and deleted
    :setup:
        1. Create groups 'group1' and group2'
        2. Start SSSD
        3. Lookup groups
    :steps:
        1. Create and add 'user1' to 'group1' and delete 'group2' and clear cache
        2. Lookup group 'group1'
        3. Lookup group 'group2'
    :expectedresults:
        1. User 'user1' is created and added to group 'group1'  and group 'group2' is deleted
        2. Group 'group1' is found and contains 'user1'
        3. Group 'group2' is not found
    :customerscenario: False
    """
    group1 = provider.group("group1").add()
    group2 = provider.group("group2").add()
    client.sssd.start()

    assert client.tools.getent.group("group1"), "Group group1 not found!"
    assert client.tools.getent.group("group2"), "Group group2 not found!"

    group1.add_member(provider.user("user1").add())
    group2.delete()

    client.sssctl.cache_expire(everything=True)

    assert client.tools.getent.group("group1"), "Group group1 not found!"
    assert "user1" in str(client.tools.getent.group("group1")), "Group group1 is missing a user!"
    assert not client.tools.getent.group("group2"), "Group group2 found!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.parametrize("obj", ["user", "group"])
@pytest.mark.parametrize("dbs", ["cache", "timestamps"])
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__invalidate_entries_in_domain_and_timestamps_caches(
    client: Client,
    provider: LDAP,
    obj: str,
    dbs: str,
):
    """
    :title: Invalidates object entries in the domain and timestamps caches using dataExpireTimestamp attribute
    :setup:
        1. Create user and group
        2. Add user as a member to group
        3. Start SSSD and lookup user, group
    :steps:
        1. Search for object attribute dataExpireTimestamp in domain and timestamps ldb caches
        2. Clear the SSSD cache for the object
        3. Search for the object attribute dataExpireTimestamp in domain and timestamps ldb caches
    :expectedresults:
        1. Attribute is found
        2. Cache is cleared for the object
        3. Attribute is found and the value is '[1]' meaning it is cleared
    :customerscenario: False
    """
    # Unable to parametrize provider method when adding objects, so all object types are created
    user = provider.user("user").add()
    provider.group("group").add().add_member(user)
    client.sssd.start()

    assert client.tools.getent.passwd("user")
    assert client.tools.getent.group("group")

    path = f"/var/lib/sss/db/{dbs}_{client.sssd.default_domain}.ldb"
    suffix = f"cn={obj}s,cn={client.sssd.default_domain},cn=sysdb"
    ldb_filter = f"dn=name={obj}@{client.sssd.default_domain},{suffix}"

    result = client.ldb.search(path, suffix, filter=ldb_filter)
    assert result != {}, f"ldbsearch {ldb_filter} did not return any results!"
    for _, v in result.items():
        result_expire_time = v.get("dataExpireTimestamp")
        assert result_expire_time is not None
        expire_time = int(result_expire_time.pop())
        assert expire_time > 1, f"Expire time for {dbs} does not have a valid value!"

    client.host.conn.exec(["sss_cache", f"-{obj[0]}", obj])

    result = client.ldb.search(path, suffix, filter=ldb_filter)
    assert result != {}, f"ldbsearch {ldb_filter} did not find any results in {path}"
    for _, v in result.items():
        result_expire_time = v.get("dataExpireTimestamp")
        assert result_expire_time is not None
        expire_time = int(result_expire_time.pop())
        assert expire_time == 1, f"Expire time for {dbs} has not been cleared!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__extra_attributes_are_stored(client: Client, provider: LDAP):
    """
    :title: Extra attributes are cached
    :setup:
        1. Create user "user1"
        2. Edit SSSD configuration and set "ldap_user_extra_attrs =
            description:gecos, userID:uidNumber, shell:loginShell, groupID:gidNumber" and
            "ldap_id_mapping = false"
        3. Start SSSD
    :steps:
        1. Lookup user
        2. Lookup user in cache
    :expectedresults:
        1. User is found
        2. User is found and cache contains correct attributes and values
    :customerscenario: True
    """
    provider.user("user1").add(gid=111111, uid=100110, gecos="gecos user1", shell="/bin/sh", home="/home/user1")
    client.sssd.domain["ldap_user_extra_attrs"] = (
        "description:gecos, userID:uidNumber, shell:loginShell, groupID:gidNumber"
    )
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.getent.passwd("user1")
    assert result is not None, "User not found!"

    search = client.ldb.search(
        f"/var/lib/sss/db/cache_{client.sssd.default_domain}.ldb", f"cn=users,cn={client.sssd.default_domain},cn=sysdb"
    )

    user_dict = search["name=user1@test,cn=users,cn=test,cn=sysdb"]
    assert user_dict["description"] == ["gecos user1"], "attribute 'description' was not correct"
    assert user_dict["shell"] == ["/bin/sh"], "attribute 'shell' was not correct"
    assert user_dict["userID"] == ["100110"], "attribute 'userID' was not correct"
    assert user_dict["groupID"] == ["111111"], "attribute 'groupID' was not correct"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__extra_attributes_with_empty_values_are_ignored(client: Client, provider: LDAP):
    """
    :title: When extra attribute of user is added but not assigned, it is neither cached nor displayed
    :setup:
        1. Create user "user1"
        2. Configure SSSD with "ldap_user_extra_attr = number:telephonenumber"
        3. Start SSSD
    :steps:
        1. Lookup user
        2. Lookup user in cache
    :expectedresults:
        1. User is found
        2. User is found and does not have the extra numbers attribute
    :customerscenario: False
    """
    provider.user("user1").add()
    client.sssd.domain["ldap_user_extra_attrs"] = "number:telephonenumber"
    client.sssd.start()

    result = client.tools.getent.passwd("user1")
    assert result is not None, "User is not found!"

    search = client.ldb.search(
        f"/var/lib/sss/db/cache_{client.sssd.default_domain}.ldb", f"cn=users,cn={client.sssd.default_domain},cn=sysdb"
    )
    assert search != {}, "User not found!"

    search = client.ldb.search(f"/var/lib/sss/db/cache_{client.sssd.default_domain}.ldb", "number=*")
    assert search == {}


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_cache__both_ldap_user_email_and_extra_attribute_email_are_stored(client: Client, ldap: LDAP):
    """
    :title: Setting ldap_user_email and email using extra attributes are cached
    :setup:
        1. Create user "user1" with gecos and mail attributes`
        2. Configure SSSD with "ldap_user_extra_attrs = email:mail, description:gecos" and
            "ldap_user_email = mail"
        3. Start SSSD
    :steps:
        1. Lookup user
        2. Lookup user in cache
    :expectedresults:
        1. User is found
        2. User is found with description, mail and email attributes
    :customerscenario: False
    """
    ldap.user("user1").add(gecos="gecos1", mail="user1@example.test")

    client.sssd.domain["ldap_user_email"] = "mail"
    client.sssd.domain["ldap_user_extra_attrs"] = "email:mail, description:gecos"
    client.sssd.start()

    result = client.tools.getent.passwd("user1")
    assert result is not None, "User is not found"
    assert result.name == "user1", "User has wrong name"

    search = client.ldb.search(
        f"/var/lib/sss/db/cache_{client.sssd.default_domain}.ldb", f"cn=users,cn={client.sssd.default_domain},cn=sysdb"
    )

    user_dict = search["name=user1@test,cn=users,cn=test,cn=sysdb"]
    assert user_dict["description"] == ["gecos1"], "attribute 'description' was not correct"
    assert user_dict["mail"] == ["user1@example.test"], "attribute 'mail' was not correct"
    assert user_dict["email"] == ["user1@example.test"], "attribute 'email' was not correct"



import subprocess

import pytest

from sssd.testlib.common.utils import sssdTools


@pytest.mark.tier1
@pytest.mark.admultiforest
class TestADMultiForest(object):

    @staticmethod
    def test_0001_multiforest(multihost, newhostname, adjoin):
        """
        :title: IDM-SSSD-TC: ad_provider: admultiforest : Authentication against two forests
        :id: 900f2467-1aca-430c-bbaa-b22d30a829ad
        :setup:
          1. Configure two domain controllers in different forests
          2. Join client to the first domain
          3. Update sssd.conf for second domain
          4. Update krb5.conf for second domain
          5. Create krb principal and keytab for second domain in the second forest and update sssd.conf
        :steps:
          1. Lookup user in the root domain of the first forest
          2. Lookup user in the  root domain fo the second forest
        :expectedresults:
          1. User is found in the root domain in the first forest
          2. User is found in the root domain in the second forest
        :customerscenario: True
        """
        adjoin(membersw='adcli')
        ad_domain = multihost.ad[0].domainname
        ad_server = multihost.ad[0].hostname
        # The second forest domain must be the last entry in the metadata file
        ad1_ip = multihost.ad[len(multihost.ad) - 1].ip
        ad1_domain = multihost.ad[len(multihost.ad) - 1].domainname
        ad1_domain_upper = str.upper(ad1_domain)
        ad1_server = multihost.ad[len(multihost.ad) - 1].hostname
        ad1_password = multihost.ad[len(multihost.ad) - 1].ssh_password

        hosts = multihost.client[0].get_file_contents('/etc/hosts', encoding='utf-8')
        hosts_new = hosts + f'{ad1_ip}    {ad1_server}'
        multihost.client[0].put_file_contents('/etc/hosts.bak', hosts)
        multihost.client[0].put_file_contents('/etc/hosts', hosts_new)

        get_keytab = f'adcli join -v -K /etc/krb5-domain1.keytab -D {ad1_domain} -S {ad1_server} --stdin-password'
        change_context = 'chcon -t krb5_keytab_t /etc/krb5-domain1.keytab'
        cleanup_krb5 = 'rm -rf /etc/krb5-domain1.keytab'
        krb5 = multihost.client[0].get_file_contents('/etc/krb5.conf', encoding='utf-8')
        substr = 'domain_realm]\n'
        new_lines = f"{ad1_domain} = {ad1_domain_upper}\n.{ad1_domain} = {ad1_domain_upper}\n"
        krb5_new = krb5.replace(substr, substr + new_lines)
        try:
            multihost.client[0].run_command(get_keytab, stdin_text=ad1_password)
        except subprocess.CalledProcessError:
            pytest.fail("adcli join failed")
        multihost.client[0].put_file_contents('/etc/krb5.conf', krb5_new)
        multihost.client[0].run_command(change_context)

        # Configure sssd
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        client.backup_sssd_conf()
        client.sssd_conf("sssd", {'domains': f'{ad_domain}, {ad1_domain}'}, action='update')
        domain_params = {
            'ad_domain': f'{ad_domain}',
            'dns_discovery_domain': f'{ad_domain}',
            'ad_server': f'{ad_server}',
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'cache_credentials': 'True',
            'dyndns_update': 'True'}
        client.sssd_conf(f'domain/{ad_domain}', domain_params, action='update')
        domain1_params = {
            'ad_domain': f'{ad1_domain}',
            'ad_server': f'{ad1_server}',
            'krb5_realm': f'{ad1_domain_upper}',
            'debug_level': '9',
            'use_fully_qualified_names': 'False',
            'cache_credentials': 'True',
            'realmd_tags': 'manages-system joined-with-samba',
            'dyndns_update': 'False',
            'krb5_keytab': '/etc/krb5-domain1.keytab',
            'ldap_krb5_keytab': '/etc/krb5-domain1.keytab',
            'id_provider': 'ad',
            'access_provider': 'ad',
            'timeout': '3600',
            'krb5_store_password_if_offline': 'True',
            'default_shell': '/bin/bash',
            'ldap_id_mapping': 'True'}
        client.sssd_conf(f'domain/{ad1_domain}', domain1_params, action='update')
        client.clear_sssd_cache()
        multihost.client[0].service_sssd('start')
        # Search for the user in same forest and domain
        getent_domain_user1 = multihost.client[0].run_command(f'getent passwd user1@{ad_domain}', raiseonerr=False)
        getent_domain_user2 = multihost.client[0].run_command(f'getent passwd user2@{ad_domain}', raiseonerr=False)
        id_domain_user1 = multihost.client[0].run_command(f'id user1@{ad_domain}', raiseonerr=False)
        id_domain_user2 = multihost.client[0].run_command(f'id user2@{ad_domain}', raiseonerr=False)
        # Search for the user in a different forest and domain
        getent_domain1_user1 = multihost.client[0].run_command(f'getent passwd user1@{ad1_domain}', raiseonerr=False)
        getent_domain1_user2 = multihost.client[0].run_command(f'getent passwd user2@{ad1_domain}', raiseonerr=False)
        id_domain1_user1 = multihost.client[0].run_command(f'id user1@{ad1_domain}', raiseonerr=False)
        id_domain1_user2 = multihost.client[0].run_command(f'id user2@{ad1_domain}', raiseonerr=False)

        multihost.client[0].put_file_contents('/etc/hosts', hosts)
        multihost.client[0].put_file_contents('/etc/krb5.conf', krb5)
        multihost.client[0].run_command(cleanup_krb5)
        client.restore_sssd_conf()

        assert getent_domain_user1.returncode == 0, f"Could not find user1 {getent_domain_user1}!"
        assert getent_domain_user2.returncode == 0, f"Could not find user1 {getent_domain_user2}!"
        assert id_domain_user1.returncode == 0, f"Could not find user1 {id_domain1_user1}!"
        assert id_domain_user2.returncode == 0, f"Could not find user2 {id_domain_user2}!"
        assert getent_domain1_user1.returncode == 0, f"Could not find user1 {getent_domain1_user1}!"
        assert getent_domain1_user2.returncode == 0, f"Could not find user2 {getent_domain1_user2}!"
        assert id_domain1_user1.returncode == 0, f"Could not find user1 {id_domain1_user1}!"
        assert id_domain1_user2.returncode == 0, f"Could not find user2 {id_domain1_user2}!"



import re
import logging
from sssd import util
from sssd.parser import SubparsersAction
from sssd.parser import Option

logger = logging.getLogger()


class RequestAnalyzer:
    """
    A request analyzer module, handles request tracking logic
    and analysis. Parses input generated from a source Reader.
    """
    module_parser = None
    consumed_logs = []
    list_opts = [
        Option('--verbose', 'Verbose output', bool, '-v'),
        Option('--pam', 'Filter only PAM requests', bool),
    ]

    show_opts = [
        Option('cid', 'Track request with this ID', int),
        Option('--child', 'Include child process logs', bool),
        Option('--merge', 'Merge logs together sorted by timestamp', bool),
        Option('--pam', 'Track only PAM requests', bool),
    ]

    def print_module_help(self, args):
        """
        Print the module parser help output

        Args:
            args (Namespace): argparse parsed arguments
        """
        self.module_parser.print_help()

    def setup_args(self, parser_grp, cli):
        """
        Setup module parser, subcommands, and options
        Args:
            parser_grp (argparse.Action): Parser group to nest
               module and subcommands under
        """
        desc = "Analyze request tracking module"
        self.module_parser = parser_grp.add_parser('request',
                                                   description=desc,
                                                   help='Request tracking')

        subparser = self.module_parser.add_subparsers(title=None,
                                                      dest='subparser',
                                                      action=SubparsersAction,
                                                      metavar='COMMANDS')

        subcmd_grp = subparser.add_parser_group('Operation Modes')
        cli.add_subcommand(subcmd_grp, 'list', 'List recent requests',
                           self.list_requests, self.list_opts)
        cli.add_subcommand(subcmd_grp, 'show', 'Track individual request ID',
                           self.track_request, self.show_opts)

        self.module_parser.set_defaults(func=self.print_module_help)

        return self.module_parser

    def get_linked_ids(self, source, pattern, regex):
        """
        Retrieve list of associated REQ_TRACE ids. Filter
        only source lines by pattern, then parse out the
        linked id with the provided regex.

        Args:
            source (Reader): source Reader object
            pattern (list of str): regex pattern(s) used for finding
                 linked ids
            regex (str): regular expression used to extract linked id

        Returns:
            List of linked ids discovered
        """
        utl = util.Utils()
        linked_ids = []
        for match in utl.matched_line(source, pattern):
            id_re = re.compile(regex)
            match = id_re.search(match)
            if match:
                found = match.group(0)
                linked_ids.append(found)
        return linked_ids

    def consume_line(self, line, source, consume):
        """
        Print or consume a line, if merge cli option is provided then consume
        boolean is set to True

        Args:
            line (str): line to process
            source (Reader): source Reader object
            consume (bool): If True, line is added to consume_logs
               list, otherwise print line

        Returns:
            True if line was processed, otherwise False
        """
        found_results = True
        if consume:
            self.consumed_logs.append(line.rstrip(line[-1]))
        else:
            # files source includes newline
            if type(source).__name__ == 'Files':
                print(line, end='')
            else:
                print(line)
        return found_results

    def print_formatted_verbose(self, source):
        """
        Parse log file and print formatted verbose list_requests output

        Args:
            source (Reader): source Reader object
        """
        data = {}
        # collect cid log lines from single run through of parsing the log
        # into dictionary # (cid, ts) -> logline_output
        for line in source:
            if "CID#" not in line:
                continue

            # parse CID and ts from line, key is a tuple of (cid,ts)
            fields = line.split("[")
            # timestamp to the minute, cut off seconds, ms
            ts = fields[0][:17]
            result = re.search('CID#[0-9]*', fields[3])
            cid = result.group(0)

            # if mapping exists, append line to output. Otherwise create new mapping
            if (cid, ts) in data.keys():
                data[(cid, ts)] += line
            else:
                data[(cid, ts)] = line

        # pretty print the data
        for k, v in data.items():
            cr_done = []
            id_done = []
            for cidline in v.splitlines():
                plugin = ""
                name = ""
                id = ""

                # CR number
                fields = cidline.split("[")
                cr_field = fields[3][7:]
                cr = cr_field.split(":")[0][4:]
                # Client connected, top-level info line
                if re.search(r'\[cmd', cidline):
                    self.print_formatted(cidline)
                # CR Plugin name
                if re.search("cache_req_send", cidline):
                    plugin = cidline.split('\'')[1]
                    id_done.clear()
                    # Extract CR number
                    fields = cidline.split("[")
                    cr_field = fields[3][7:]
                    cr = cr_field.split(":")[0][4:]
                # CR Input name
                elif re.search("cache_req_process_input", cidline):
                    name = cidline.rsplit('[')[-1]
                # CR Input id
                elif re.search("cache_req_search_send", cidline):
                    id = cidline.rsplit()[-1]

                if plugin:
                    print("   - " + plugin)
                if name:
                    # Avoid duplicate output with the same CR #
                    if cr not in cr_done:
                        print("       - " + name[:-1])
                        cr_done.append(cr)
                if (id and ("UID" in cidline or "GID" in cidline)):
                    if id not in id_done and bool(re.search(r'\d', id)):
                        print("       - " + id)
                        id_done.append(id)

    def print_formatted(self, line):
        """
        Parse line and print formatted list_requests output

        Args:
            line (str): line to parse
        Returns:
            Client ID from printed line, 0 otherwise
        """
        # exclude backtrace logs
        if line.startswith('   *  '):
            return 0
        if "refreshed" in line:
            return 0
        ts = line.split(")")[0]
        ts = ts[1:]
        fields = line.split("[")
        cid = fields[3][4:-9]
        cmd = fields[4][4:-1]
        uid = fields[5][4:-1]
        if not uid.isnumeric():
            uid = fields[6][4:-1]
        print(f'{ts}: [uid {uid}] CID #{cid}: {cmd}')
        return cid

    def list_requests(self, args):
        """
        List component (default: NSS) responder requests

        Args:
            args (Namespace):  populated argparse namespace
        """
        utl = util.Utils()
        source = utl.load(args)
        component = source.Component.NSS
        resp = "nss"
        # Log messages matching the following regex patterns contain
        # the useful info we need to produce list output
        patterns = [r'\[cmd']
        if args.pam:
            component = source.Component.PAM
            resp = "pam"

        logger.info(f"******** Listing {resp} client requests ********")
        source.set_component(component, False)

        if args.verbose:
            self.print_formatted_verbose(source)
        else:
            for line in utl.matched_line(source, patterns):
                if type(source).__name__ == 'Journald':
                    print(line)
                else:
                    self.print_formatted(line)

    def track_request(self, args):
        """
        Print Logs pertaining to individual SSSD client request

        Args:
            args (Namespace):  populated argparse namespace
        """
        utl = util.Utils()
        source = utl.load(args)
        cid = args.cid
        resp_results = False
        be_results = False
        component = source.Component.NSS
        resp = "nss"
        pattern = [rf"\[CID#{cid}\]"]

        if args.pam:
            component = source.Component.PAM
            resp = "pam"

        logger.info(f"******** Checking {resp} responder for Client ID"
                    f" {cid} *******")
        source.set_component(component, args.child)
        for match in utl.matched_line(source, pattern):
            resp_results = self.consume_line(match, source, args.merge)

        logger.info(f"********* Checking Backend for Client ID {cid} ********")
        pattern = [rf'REQ_TRACE.*\[sssd.{resp} CID #{cid}\]']
        source.set_component(source.Component.BE, args.child)

        be_id_regex = r'\[RID#[0-9]+\]'
        be_ids = self.get_linked_ids(source, pattern, be_id_regex)

        pattern.clear()
        [pattern.append(f'\\{id}') for id in be_ids]

        for match in utl.matched_line(source, pattern):
            be_results = self.consume_line(match, source, args.merge)

        if args.merge:
            # sort by date/timestamp
            sorted_list = sorted(self.consumed_logs,
                                 key=lambda s: s.split(')')[0])
            for entry in sorted_list:
                print(entry)
        if not resp_results and not be_results:
            logger.warn(f"ID {cid} not found in logs!")



"""Automation for ldap_query_time_logging

:requirement: LDAP Query Time Logging
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
:bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1925559
"""
from __future__ import print_function
import re
import pytest
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.timelog
@pytest.mark.tier1_3
class TestLdapTimeLogging(object):
    """ Test ldap_time_logging """

    def test_0001_bz1925559(self, multihost, backupsssdconf):
        """
        :title: Check time logging is not enabled at debug_level 8"
        :description: Debug level of 8 and below does not enable time logging
        :id: 11379bae-99bd-4dd9-bee6-fa0dba28f4fe
        :steps:
          1. Set debug_level to 8 in sssd.conf
          2. Start sssd with cleared logs
          3. Lookup user
          4. Check logs in /var/log/sssd/
        :expectedresults:
          1. sssd should use default debug level 8
          2. sssd services start successfully
          3. Lookup user succeeds
          4. Log files has no time logged for ldap queries
        """
        section = f'domain/{ds_instance_name}'
        tools = sssdTools(multihost.client[0])
        domain_params = {'debug_level': '8'}
        tools.sssd_conf(section, domain_params)
        tools.clear_sssd_cache()

        user = f'foo9@{ds_instance_name}'
        get_user = f'getent passwd {user}'
        multihost.client[0].run_command(get_user, raiseonerr=False)
        log_list = ['sssd', f'sssd_{ds_instance_name}',
                    'sssd_nss', 'sssd_pam']
        for log in log_list:
            log = f'/var/log/sssd/{log}.log'
            log_str = multihost.client[0].get_file_contents(log).decode('utf-8')
            find = re.compile(r'milliseconds')
            assert not find.search(log_str)

    def test_0002_bz1925559(self, multihost, backupsssdconf):
        """
        :title: Time logging is enabled by bitmask 0x20000
        :description: 0x20000 is the bitmask to select the timing/statistical
         data. So time logging is only enabled when the bitmask is used directly
         or debug level is set to 9
        :id: 9fad8c34-bbac-43bd-81ae-fb1540075f2c
        :steps:
          1. Set debug_level to 9 in sssd.conf
          2. Start sssd with cleared logs
          3. Lookup user
          4. Check domain logs in /var/log/sssd/
          5. Repeat 1 - 4 with debug level 0x20000
        :expectedresults:
          1. sssd should use default debug level 9
          2. sssd services start successfully
          3. Lookup user succeeds
          4. Log files has time logged for ldap queries
          5. Should succeed
        """
        section = f'domain/{ds_instance_name}'
        tools = sssdTools(multihost.client[0])

        # Test with debug level 9
        domain_params = {'debug_level': '9'}
        tools.sssd_conf(section, domain_params)
        tools.clear_sssd_cache()
        user = f'foo3@{ds_instance_name}'
        get_user = f'getent passwd {user}'
        multihost.client[0].run_command(get_user, raiseonerr=False)
        log = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        log_str1 = multihost.client[0].get_file_contents(log).decode('utf-8')

        # Test with debug level 0x20000
        domain_params = {'debug_level': '0x20000'}
        tools.sssd_conf(section, domain_params)
        tools.clear_sssd_cache()
        multihost.client[0].run_command(get_user, raiseonerr=False)
        log_str2 = multihost.client[0].get_file_contents(log).decode('utf-8')

        find1 = re.compile(r'sdap_call_op_callback.*\[\d+[.]\d+\] milliseconds')
        find2 = re.compile(r'dp_req_done.*\[\d+[.]\d+\] milliseconds')
        assert find1.search(log_str1)
        assert find2.search(log_str1)
        assert find1.search(log_str2)
        assert find2.search(log_str2)

    def test_0003_bz1925559(self, multihost, backupsssdconf):
        """
        :title: Warning message for long queries without debug level set
        :id: 9280b9fa-e0b4-446a-92ad-258888195633
        :steps:
          1. Remove debug_level to enable default debug level
          2. Set delay to 800ms using tc on all interfaces of the client
          3. Set ldap_search_timeout to 1 second
          4. Start sssd with cleared logs
          5. Lookup user
          6. Check domain logs in /var/log/sssd/
        :expectedresults:
          1. sssd should use default debug level
          2. tc configures interfaces sucessfully
          3. Should succeed
          4. sssd services start successfully
          5. Lookup user succeeds with delay
          6. Domain log has a warning message for long query time
        """
        section = f'domain/{ds_instance_name}'
        multihost.client[0].run_command("modprobe sch_netem")
        tools = sssdTools(multihost.client[0])
        domain_params = {'debug_level': ''}
        tools.sssd_conf(section, domain_params, action="delete")
        domain_params = {'ldap_search_timeout': '1'}
        tools.sssd_conf(section, domain_params)

        get_intf = "nmcli device status | grep connected | awk '{print $1}'"
        intf = multihost.client[0].run_command(get_intf, raiseonerr=True)
        intf_list = intf.stdout_text.splitlines()
        print(f'List of interfaces - {intf_list}')
        pkgs = 'yum install -y iproute-tc kernel-modules-extra'
        multihost.client[0].run_command(pkgs, raiseonerr=True)
        for interface in intf_list:
            tc_rule = f'tc qdisc add dev {interface} root netem delay 800ms'
            multihost.client[0].run_command(tc_rule, raiseonerr=True)
            get_tc_rule = f'tc qdisc show dev {interface}'
            show_rule = multihost.client[0].run_command(get_tc_rule, raiseonerr=True)
            print(f'===tc rules for {interface}===:\n{show_rule.stdout_text}')
        tools.clear_sssd_cache()
        user = f'foo5@{ds_instance_name}'
        get_user = f'getent passwd {user}'
        multihost.client[0].run_command(get_user, raiseonerr=False)
        log = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        log_str = multihost.client[0].get_file_contents(log).decode('utf-8')
        for interface in intf_list:
            tc_rule = f'tc qdisc del dev {interface} root'
            multihost.client[0].run_command(tc_rule, raiseonerr=True)
        print(log_str)
        find = re.compile(r'sdap_call_op_callback.*more than 80% of timeout')
        assert find.search(log_str)
